
from typing import Dict, Optional, Union, Tuple, List
from . import DataManager
import pandas as pd
import numpy as np
import datetime
import time
import io
import copy
import json
import contextlib
import re
from ..wrapper.mysql import RawDatabaseConnector, BasicDatabaseConnector, DerivedDatabaseConnector
from .fund_info_filter import filter_fund_info, fund_info_update, active_fund_info, get_ipos_funds, get_conv_funds
from ..struct import AssetPrice, FundScoreParam, TaaTunerParam, AssetTimeSpan
from .score import FundScoreManager, ScoreFunc
from .data_tables import FundDataTables
from .etl_tool import S3RetrieverForDM
from ..view.raw_models import CSIndexComponent
from ..view.basic_models import *
from ..view.derived_models import *

INDEX_REPLACE = ('active','csi_stockfund')

@contextlib.contextmanager
def profiled(file_name=None):
    import cProfile
    import pstats

    pr = cProfile.Profile()
    pr.enable()
    yield
    pr.disable()
    s = io.StringIO()
    ps = pstats.Stats(pr, stream=s).sort_stats('cumulative')
    ps.print_stats()
    # uncomment this to see who's calling what
    # ps.print_callers()
    if file_name is None:
        print(s.getvalue())
    else:
        open(file_name, 'w').write(s.getvalue())


class FundDataManager(DataManager):

    _BLOCK_DF_LIST = ('fund_manager_score', 'fund_indicator')

    def __init__(self, start_time=None, end_time=None, fund_score_param: FundScoreParam = None, score_manager: FundScoreManager = None, s3_retriever_activation: bool = False):
        DataManager.__init__(self,
            start_time or datetime.datetime(2005, 1, 1),
            end_time or datetime.datetime.now()
        )
        self._s3_retriever = S3RetrieverForDM(activation=s3_retriever_activation)
        self._fund_score_param = fund_score_param or FundScoreParam(tag_type=1, score_method=1, is_full=1)
        self.dts = FundDataTables()
        self._is_inited = False
        self._score_manager = score_manager or FundScoreManager()
        self.set_score_param(self._fund_score_param)
        self.fund_list_with_nav = set()

    @property
    def inited(self):
        return self._is_inited

    def _init_from_s3_part(self) -> bool:
        # 获取所有dts
        failed_tasks: List[str] = self._s3_retriever.retrieve_all_dfs(self)
        if failed_tasks:
            print(f'retrieve dm failed from s3 (err_info){failed_tasks}')
            return False

        # 其余初始化流程
        self.fund_list_with_nav = set(self.dts.fund_nav.columns.union(self.dts.fund_unit_nav.columns).array)
        self.end_date_dic = self.dts.fund_info[['fund_id', 'end_date']].set_index('fund_id').to_dict()['end_date']
        self.dts.fix_data()
        self._score_manager.set_dts(self.dts)
        return True

    # 使用全默认配置的DM(包括DM自己的参数以及FundScoreManager等)
    def _init_from_s3_with_all_default_config(self):
        print('try to init from s3 with all default config')

        # 初始化取得默认配置的DM
        if not self._init_from_s3_part():
            return

        # 取得默认算好的fund score
        failed_tasks: List[str] = self._s3_retriever.retrieve_fund_score(self._score_manager)
        if failed_tasks:
            print(f'retrieve fund score for dm from s3 failed (err_info){failed_tasks}')
            return

        self._is_inited = True
        print('init from s3 with all default config done')

    # 使用默认配置的DM+自定义的FundScoreManager
    def _init_from_s3_with_score_customized(self, score_pre_calc, use_weekly_monthly_indicators):
        print('try to init from s3 with score customized')

        # 初始化取得默认配置的DM
        if not self._init_from_s3_part():
            return

        _t0 = time.time()
        if score_pre_calc:
            # 由于FundScoreManager是自定义的，这里重新算一下
            self._score_manager.pre_calculate(is_filter_c=True, use_weekly_monthly_indicators=use_weekly_monthly_indicators)
            self._score_manager.pre_calculate_manager()
        _t1 = time.time()

        self._is_inited = True
        print(f'init from s3 with score customized done, score pre calculation costs {_t1 - _t0}')

    def try_to_init_from_s3(self, default_value_flag: List[bool], score_pre_calc, use_weekly_monthly_indicators, block_df_list) -> bool:
        # 判断init中的参数是否都是与默认值相同的值
        if not all(default_value_flag):
            return False

        # 判断FundScoreManager相关的参数是否是默认的
        score_manager = FundScoreManager()
        score_manager.set_param(FundScoreParam(tag_type=1, score_method=1, is_full=1))
        if self._score_manager == score_manager:
            self._init_from_s3_with_all_default_config()
        else:
            self._init_from_s3_with_score_customized(score_pre_calc, use_weekly_monthly_indicators)
        if not self._is_inited:
            print('init failed from s3')
            return False

        # 干掉不用的字段以节省内存
        self.dts.remove_fields(block_df_list)
        return True

    def init(self, index_list=None, score_pre_calc=True, print_time=False, use_weekly_monthly_indicators=False, block_df_list=_BLOCK_DF_LIST):
        if self._s3_retriever.is_activated:
            # Caution!! 这里应始终保持init传进来的参数与try_to_init_from_s3第一个参数相适应，以正确判断是否为默认配置
            # 如果init中的参数改成keyword类型，可以使用__kwdefaults__来帮助判断
            std_dm = DataManager(datetime.datetime(2005, 1, 1), datetime.datetime.now())
            if self.try_to_init_from_s3([self.start_date == std_dm.start_date,
                                        self.end_date == std_dm.end_date,
                                        index_list is None or not index_list,
                                        use_weekly_monthly_indicators is False,
                                        ], score_pre_calc, use_weekly_monthly_indicators, block_df_list):
                # 从s3成功读取数据，直接返回
                return
        print('could not init from s3 directly, do normal initialization')
        _tm_start = time.time()
        index_list = index_list or list(AssetTimeSpan.__dataclass_fields__.keys())
        index_list = [_.replace(INDEX_REPLACE[0],INDEX_REPLACE[1]) for _ in index_list]

        def fetch_table(session, view):
            query = session.query(view)
            return self._s3_retriever.fetch_table(query)

        def _index_price_filter(query, index_list):
            def _index_price_filter_etl(x):
                return x[x.index_id.isin(index_list) & (x.datetime >= self.start_date) & (x.datetime <= self.end_date)].reset_index(drop=True)

            return query.filter(
                IndexPrice.index_id.in_(index_list),
                IndexPrice.datetime >= self.start_date,
                IndexPrice.datetime <= self.end_date
            ), _index_price_filter_etl

        def _fund_nav_filter(query, fund_id_list):
            def _fund_nav_filter_etl(x):
                return x[x.fund_id.isin(fund_id_list) & (x.datetime >= self.start_date) & (x.datetime <= self.end_date)].reset_index(drop=True)

            return query.filter(
                FundNav.fund_id.in_(fund_id_list),
                FundNav.datetime >= self.start_date,
                FundNav.datetime <= self.end_date,
            ), _fund_nav_filter_etl

        def _fund_size_and_hold_rate_filter(query, fund_id_list):
            def _fund_size_and_hold_rate_filter_etl(x):
                return x[x.fund_id.isin(fund_id_list) & (x.datetime >= self.start_date - datetime.timedelta(days=200)) & (x.datetime <= self.end_date)].reset_index(drop=True)

            return query.filter(
                Fund_size_and_hold_rate.fund_id.in_(fund_id_list),
                Fund_size_and_hold_rate.datetime >= self.start_date - datetime.timedelta(days=200),
                Fund_size_and_hold_rate.datetime <= self.end_date,
            ), _fund_size_and_hold_rate_filter_etl

        def _fund_indicator_filter(query, fund_id_list):
            def _fund_indicator_filter_etl(x):
                return x[x.fund_id.isin(fund_id_list) & (x.datetime >= self.start_date) & (x.datetime <= self.end_date)].reset_index(drop=True)

            return query.filter(
                FundIndicator.fund_id.in_(fund_id_list),
                FundIndicator.datetime >= self.start_date,
                FundIndicator.datetime <= self.end_date,
            ), _fund_indicator_filter_etl

        def _fund_indicator_annual_filter(query):
            def _fund_indicator_annual_filter_etl(x):
                return x[x.fund_id.isin(self.dts.fund_list) & (x.datetime <= self.end_date)].reset_index(drop=True)

            return query.filter(
                FundIndicatorAnnual.fund_id.in_(self.dts.fund_list),
                FundIndicatorAnnual.datetime <= self.end_date,
            ), _fund_indicator_annual_filter_etl

        def _index_valuation_long_term_filter(query, pct_index_list):
            def _index_valuation_long_term_filter_etl(x):
                x = x[x.index_id.isin(pct_index_list) & (x.datetime >= self.start_date) & (x.datetime <= self.end_date)].reset_index(drop=True)
                return x.sort_values(by='datetime', axis=0)

            return query.filter(
                IndexValuationLongTerm.index_id.in_(pct_index_list),
                IndexValuationLongTerm.datetime >= self.start_date,
                IndexValuationLongTerm.datetime <= self.end_date,
            ).order_by(IndexValuationLongTerm.datetime.asc()), _index_valuation_long_term_filter_etl

        def _fund_indicator_weekly_filter(query):
            def _fund_indicator_weekly_filter_etl(x):
                return x[x.fund_id.isin(self.dts.fund_list) & (x.datetime >= self.start_date) & (x.datetime <= self.end_date)].reset_index(drop=True)

            return query.filter(
                FundIndicatorWeekly.fund_id.in_(self.dts.fund_list),
                FundIndicatorWeekly.datetime >= self.start_date,
                FundIndicatorWeekly.datetime <= self.end_date,
            ), _fund_indicator_weekly_filter_etl

        def _fund_manager_score_filter(query):
            def _fund_manager_score_filter_etl(x):
                return x[x.fund_type.isin(['stock']) & (x.datetime >= self.start_date) & (x.datetime <= self.end_date)].reset_index(drop=True)

            return query.filter(
                FundManagerScore.fund_type.in_(['stock']),
                FundManagerScore.datetime >= self.start_date,
                FundManagerScore.datetime <= self.end_date,
            ), _fund_manager_score_filter_etl

        def _fund_manager_fund_rank_info_filter(query):
            def _fund_manager_fund_rank_info_filter_etl(x):
                return x[(x.datetime >= self.start_date) & (x.datetime <= self.end_date)].reset_index(drop=True)

            return query.filter(
                FundManagerFundRank.datetime >= self.start_date,
                FundManagerFundRank.datetime <= self.end_date,
            ), _fund_manager_fund_rank_info_filter_etl

        with RawDatabaseConnector().managed_session() as raw_session:
            _tm_raw_start = time.time()
            self.dts.cs_index_component = fetch_table(raw_session, CSIndexComponent)
            _tm_raw_cs_index_component = time.time()
            if print_time:
                print(f'\t[time][raw] fetch cs index component: {_tm_raw_cs_index_component - _tm_raw_start}')

        # info is necessary
        with BasicDatabaseConnector().managed_session() as quant_session:
            _tm_basic_start = time.time()
            self.dts.fund_info = fetch_table(quant_session, FundInfo)
            self.dts.fund_benchmark_info = fetch_table(quant_session, FundBenchmark)
            self.dts.active_fund_info = active_fund_info(self.dts.fund_info, self.dts.fund_benchmark_info)
            _tm_basic_fetch_info = time.time()
            self.dts.fund_info = fund_info_update(self.dts.fund_info)
            self.dts.fund_status_latest = fetch_table(quant_session, FundStatusLatest)
            _fund_open_info = fetch_table(quant_session, FundOpenInfo)
            _filter_fund_info = filter_fund_info(self.dts.fund_info, index_list)
            _tm_basic_filter_info = time.time()
            self.dts.fund_list = set(_filter_fund_info.fund_id)
            self.dts.all_fund_list = self.dts.fund_list.union(self.dts.active_fund_info.fund_id)
            in_market_fund_list = self.dts.fund_status_latest[~self.dts.fund_status_latest['trade_status'].isnull()].fund_id.tolist()
            close_fund_list = _fund_open_info.fund_id.tolist()
            self.dts.fund_ipo_stats = fetch_table(quant_session, FundIPOStats)
            self.dts.fund_conv_stats = fetch_table(quant_session, FundConvStats)
            self.dts.fund_ipo_list = get_ipos_funds(self.dts.fund_ipo_stats)
            self.dts.fund_conv_list = get_conv_funds(self.dts.fund_conv_stats)
            self.dts.fund_ipo_list = {i for i in self.dts.fund_ipo_list if (i not in in_market_fund_list) and (i not in close_fund_list)}
            self.dts.fund_conv_list = {i for i in self.dts.fund_conv_list if (i not in in_market_fund_list) and (i not in close_fund_list)}
            self.dts.all_fund_list = {i for i in self.dts.all_fund_list if (i not in in_market_fund_list) and (i not in close_fund_list)}
            self.dts.all_fund_list = self.dts.all_fund_list.union(self.dts.fund_ipo_list).union(self.dts.fund_conv_list)
            _normal_funds_index_dic = {cur.fund_id: cur.index_id for cur in _filter_fund_info.itertuples()}
            _active_funds_index_dic = {i.fund_id: i.index_id for i in self.dts.active_fund_info.itertuples()}
            self.dts.fund_index_map = {**_normal_funds_index_dic, **_active_funds_index_dic}
            self.end_date_dic = self.dts.fund_info[['fund_id', 'end_date']].set_index('fund_id').to_dict()['end_date']
            _tm_basic_prep_struct = time.time()
            self.dts.trading_days = fetch_table(quant_session, TradingDayList)
            self.dts.trading_days = self.dts.trading_days[self.dts.trading_days.datetime <= self.end_date]
            trim_trading_days = self.dts.trading_days[(self.dts.trading_days.datetime >= self.start_date) & (self.dts.trading_days.datetime <= self.end_date)].datetime
            _tm_basic_fetch_days = time.time()
            # index
            self.dts.index_info = fetch_table(quant_session, IndexInfo)
            _tm_basic_fetch_index_info = time.time()
            _index_query = quant_session.query(
                IndexPrice.index_id,
                IndexPrice.datetime,
                IndexPrice.close
            )
            _index_query, _index_price_filter_func = _index_price_filter(_index_query, index_list)
            index_price_raw = self._s3_retriever.fetch_table(_index_query, _index_price_filter_func)
            self.dts.index_price = index_price_raw.pivot_table(index='datetime', columns='index_id', values='close').reindex(trim_trading_days).fillna(method='ffill').rename(columns = {INDEX_REPLACE[1]:INDEX_REPLACE[0]})
            _tm_basic_fetch_index_price = time.time()
            # fund nav
            _fund_nav_query = quant_session.query(
                FundNav.fund_id,
                FundNav.adjusted_net_value,
                FundNav.unit_net_value,
                FundNav.datetime
                # FundNav.subscribe_status,
                # FundNav.redeem_status,
            )
            _fund_nav_query, _fund_nav_filter_func = _fund_nav_filter(_fund_nav_query,self.dts.all_fund_list)
            _nav_df = self._s3_retriever.fetch_table(_fund_nav_query, _fund_nav_filter_func)
            _tm_basic_fetch_fund_nav = time.time()

            # 这里我们制作一个总的pivot_table再取值，会比分别两次pivot_table快一些
            # 如果index-column pair有重复的值，pivot会报错而pivot_table不会，所以这里用pivot也许更好些
            # 但是pivot里边会调用set_index，导致datetime列的dtype从object变为datetime64[ns]
            _temp_fund_nav = _nav_df.pivot_table(index='datetime', columns='fund_id')
            self.dts.fund_nav = _temp_fund_nav['adjusted_net_value'].dropna(axis=1, how='all').fillna(method='ffill')
            self.dts.fund_unit_nav = _temp_fund_nav['unit_net_value'].dropna(axis=1, how='all').fillna(method='ffill')
            self.fund_list_with_nav = set(self.dts.fund_nav.columns.union(self.dts.fund_unit_nav.columns).array)
            _tm_basic_pivot_fund_nav = time.time()

            # fund size and fund com hold
            # 机构持仓比例每六个月更新，提前取半年以上, 填充空白值用
            # fund_com_hold fund_size 和 fund_nav同结构，日期纵轴，基金代码横轴
            _fund_size_and_hold_rate_query = quant_session.query(
                Fund_size_and_hold_rate.fund_id,
                Fund_size_and_hold_rate.size,
                Fund_size_and_hold_rate.institution_holds,
                Fund_size_and_hold_rate.datetime
            )
            _fund_size_and_hold_rate_query, _fund_size_and_hold_rate_filter_func = _fund_size_and_hold_rate_filter(_fund_size_and_hold_rate_query,self.dts.all_fund_list)
            _size_and_hold_rate = self._s3_retriever.fetch_table(_fund_size_and_hold_rate_query, _fund_size_and_hold_rate_filter_func)
            _dt_index = self.dts.trading_days[self.dts.trading_days.datetime >= (self.start_date - datetime.timedelta(days=200))].datetime
            _temp_size_and_hold_rate = _size_and_hold_rate.pivot(index='datetime', columns='fund_id').reindex(_dt_index).fillna(method='ffill')
            self.dts.fund_size = _temp_size_and_hold_rate['size']
            self.dts.fund_com_hold = _temp_size_and_hold_rate['institution_holds']
            _tm_basic_fetch_fund_size = time.time()
            if print_time:
                print(f'\t[time][basic] fetch fund info : {_tm_basic_fetch_info - _tm_basic_start}')
                print(f'\t[time][basic] filter fund info: {_tm_basic_filter_info - _tm_basic_fetch_info}')
                print(f'\t[time][basic] prep fund struct: {_tm_basic_prep_struct - _tm_basic_filter_info}')
                print(f'\t[time][basic] fetch trade days: {_tm_basic_fetch_days - _tm_basic_prep_struct}')
                print(f'\t[time][basic] fetch index info: {_tm_basic_fetch_index_info - _tm_basic_fetch_days}')
                print(f'\t[time][basic] fetch index price: {_tm_basic_fetch_index_price - _tm_basic_fetch_index_info}')
                print(f'\t[time][basic] fetch fund navs: {_tm_basic_fetch_fund_nav - _tm_basic_fetch_index_price}')
                print(f'\t[time][basic] pivot fund navs: {_tm_basic_pivot_fund_nav - _tm_basic_fetch_fund_nav}')
                print(f'\t[time][basic] fetch fund size: {_tm_basic_fetch_fund_size - _tm_basic_pivot_fund_nav}')
        _tm_basic = time.time()

        # get derived data first
        with DerivedDatabaseConnector().managed_session() as derived_session:
            _tm_deriv_start = time.time()
            _fund_indicator_query = derived_session.query(
                FundIndicator.fund_id,
                FundIndicator.datetime,
                FundIndicator.alpha,
                FundIndicator.beta,
                FundIndicator.fee_rate,
                FundIndicator.track_err,
                FundIndicator.year_length,
                FundIndicator.info_ratio,
                # FundIndicator.down_risk,
                # FundIndicator.ret_over_period,
                # FundIndicator.treynor,
                # FundIndicator.mdd,
                # FundIndicator.m_square,
                # FundIndicator.var,
                # FundIndicator.r_square,
                # FundIndicator.sharpe,
                # FundIndicator.annual_ret,
                # FundIndicator.annual_vol,
                # FundIndicator.time_ret,
            )
            _fund_indicator_query, _fund_indicator_filter_func = _fund_indicator_filter(_fund_indicator_query, self.dts.all_fund_list)
            self.dts.fund_indicator = self._s3_retriever.fetch_table(_fund_indicator_query, _fund_indicator_filter_func)
            _tm_deriv_fetch_fund_indicator = time.time()
            _tm_deriv_apply_fund_indicator = time.time()
            _fund_indicator_query = derived_session.query(
                FundIndicatorAnnual
            )
            _fund_indicator_query, _fund_indicator_filter_func = _fund_indicator_annual_filter(_fund_indicator_query)
            self.dts.fund_indicator_annual = self._s3_retriever.fetch_table(_fund_indicator_query, _fund_indicator_filter_func)
            self.dts.fund_indicator_annual = self.dts.fund_indicator_annual.set_index(['fund_id','datetime'])
            _tm_deriv_fetch_fund_indicator_annual = time.time()
            # fetch index val data from derived table, and choose different pct value for different index_id
            _tm_deriv_fetch_index_pct = time.time()
            pct_index_list = list(TaaTunerParam.POOL.keys())
            _index_pct_query = derived_session.query(
                IndexValuationLongTerm
            )
            _index_pct_query, _index_pct_filter_func = _index_valuation_long_term_filter(_index_pct_query, pct_index_list)
            self.dts._index_pct_df = self._s3_retriever.fetch_table(_index_pct_query, _index_pct_filter_func).set_index('index_id')
            # 按照trading date 作为index, fillna
            date_list = self.dts.trading_days.datetime.tolist()
            res = []
            for index_id in pct_index_list:
                df = self.dts._index_pct_df[self.dts._index_pct_df.index == index_id].set_index('datetime').reindex(date_list).reset_index().fillna(method='ffill')
                df.loc[:, 'index_id'] = index_id
                res.append(df)
            _index_pct_df = pd.concat(res, axis=0)
            self.dts.index_pct = _index_pct_df[['datetime','index_id','pe_pct_5_3','pb_pct_5_3','ps_pct_5_3']].rename(
                columns={'pe_pct_5_3':'pe_pct','pb_pct_5_3':'pb_pct','ps_pct_5_3':'ps_pct'}).pivot_table(index=['datetime', 'index_id'], values=['pe_pct', 'pb_pct', 'ps_pct'])
            self.dts.index_date_list = list(self.dts.index_pct.index.remove_unused_levels().levels[0])
            _tm_deriv_apply_index_pct = time.time()

            _fund_manager_score_query = derived_session.query(
                FundManagerScore.manager_id,
                FundManagerScore.datetime,
                FundManagerScore.trade_year,
                FundManagerScore.annualized_ret,
                FundManagerScore.annualized_vol,
                FundManagerScore.model_alpha,
                FundManagerScore.model_beta,
                FundManagerScore.sharpe,
                FundManagerScore.fund_type,
            )
            _fund_manager_score_query, _fund_manager_score_filter_func = _fund_manager_score_filter(_fund_manager_score_query)
            self.dts.fund_manager_score = self._s3_retriever.fetch_table(_fund_manager_score_query, _fund_manager_score_filter_func).drop(columns='fund_type')
            self.dts.fund_manager_score = self.dts.fund_manager_score.pivot_table(index='datetime', columns='manager_id').reindex(trim_trading_days).stack().reset_index()
            _tm_deriv_apply_active_indicator = time.time()
            _fund_manager_info_query = derived_session.query(
                FundManagerInfo.mng_id,
                FundManagerInfo.start_date,
                FundManagerInfo.fund_id,
                FundManagerInfo.end_date,
            )
            self.dts.fund_manager_info = self._s3_retriever.fetch_table(_fund_manager_info_query)
            self.dts.fund_manager_info = self.dts.fund_manager_info.set_index('mng_id')
            _tm_deriv_apply_manager_info = time.time()
            _fund_manager_fund_rank_info_query = derived_session.query(
                FundManagerFundRank.mng_id,
                FundManagerFundRank.datetime,
                FundManagerFundRank.fund_list,
            )
            _fund_manager_fund_rank_info_query, _fund_manager_fund_rank_info_filter_func = _fund_manager_fund_rank_info_filter(_fund_manager_fund_rank_info_query)
            self.dts.fund_manager_rank = self._s3_retriever.fetch_table(_fund_manager_fund_rank_info_query, _fund_manager_fund_rank_info_filter_func)
            self.dts.fund_manager_rank['fund_list'] = [ json.loads(i) for i in self.dts.fund_manager_rank.fund_list]
            self.dts.fund_manager_rank = self.dts.fund_manager_rank.set_index(['datetime','mng_id'])
            _fund_manager_score = self.dts.fund_manager_score.rename(columns={'manager_id':'mng_id'}).set_index(['datetime','mng_id'])
            _df = _fund_manager_score.join(self.dts.fund_manager_rank).copy()
            _df = _df.explode('fund_list').dropna()
            _df = _df.reset_index().drop('mng_id',axis=1).rename(columns={'fund_list':'fund_id'})
            _cols = _df.columns.tolist()
            _mng_score_dic = { c : f'mng_{c}' for c in _cols if c not in ['datetime', 'fund_id']}
            _df = _df.rename(columns = _mng_score_dic)
            _df['datetime'] = _df.datetime.dt.date
            self.dts.fund_indicator = pd.merge(self.dts.fund_indicator, _df,how='left',left_on=['fund_id','datetime'],right_on=['fund_id','datetime'])
            #self.dts.fund_indicator['active'] = self.dts.fund_indicator['active'].fillna(0)
            _index_list = ['hs300','csi500','gem']
            
            self.dts.fund_indicator['active'] = self.dts.fund_indicator.fund_id.transform(lambda x: 0 if x in _normal_funds_index_dic and _normal_funds_index_dic[x] in _index_list else 1)
            self.dts.fund_indicator['index_id'] = self.dts.fund_indicator.fund_id.apply(lambda x: self.dts.fund_index_map.get(x, None))
            self.dts.fund_indicator = self.dts.fund_indicator.dropna(subset=['index_id'])
            _tm_deriv_apply_manager_rank = time.time()

            if use_weekly_monthly_indicators:
                # indicator weekly 只有周日有值 在每一个fund上fillna, merge week_ly into indicator
                _fund_indicator_weekly_query = derived_session.query(
                    FundIndicatorWeekly.fund_id,
                    FundIndicatorWeekly.datetime,
                    FundIndicatorWeekly.alpha_w,
                    FundIndicatorWeekly.beta_w,
                    FundIndicatorWeekly.track_err_w,
                )
                _fund_indicator_weekly_query, _fund_indicator_weekly_filter_func = _fund_indicator_weekly_filter(_fund_indicator_weekly_query)
                _fund_indicator_weekly = self._s3_retriever.fetch_table(_fund_indicator_weekly_query, _fund_indicator_weekly_filter_func)
                _fund_indicator_weekly = _fund_indicator_weekly.pivot_table(index='datetime', columns='fund_id')
                _dt = _fund_indicator_weekly.index.union(date_list).set_names('datetime')  # 交易日，和现有indicator index 周日做并集
                _fund_indicator_weekly = _fund_indicator_weekly.reindex(index=_dt).fillna(method='ffill').stack().swaplevel()
                _fund_indicator_weekly = _fund_indicator_weekly.groupby(level='fund_id').apply(lambda x: x.droplevel(level=0).loc[:self.end_date_dic[x.index.get_level_values(0).array[0]], :])
                _tm_process_indicator_weekly = time.time()

                # indicator monthly 只有每月最后一天有值 在每一个fund上fillna, merge monthly into indicator
                _fund_indicator_monthly = fetch_table(derived_session, FundIndicatorMonthly)
                _fund_indicator_monthly = _fund_indicator_monthly.pivot_table(index='datetime', columns='fund_id')
                _dt = _fund_indicator_monthly.index.union(date_list).set_names('datetime')  # 交易日，和现有indicator index做并集
                _fund_indicator_monthly = _fund_indicator_monthly.reindex(index=_dt).fillna(method='ffill').stack().swaplevel()
                _fund_indicator_monthly = _fund_indicator_monthly.groupby(level='fund_id').apply(lambda x: x.droplevel(level=0).loc[:self.end_date_dic[x.index.get_level_values(0).array[0]], :])
                _tm_process_indicator_monthly = time.time()
                # 本来join可以传一个list of df进去一起join，效率应该更高，但这里用了on参数指定fund_indicator的其中两列（而不是index），这样便只能一次传一个df进去（pandas不支持这种情况下传一个list）
                for df in [_fund_indicator_weekly, _fund_indicator_monthly]:
                    self.dts.fund_indicator = self.dts.fund_indicator.join(df, on=['fund_id', 'datetime'], how='left')
                _tm_process_indicator_join = time.time()
                self.dts.fund_indicator_annual.index.set_levels(self.dts.fund_indicator_annual.index.levels[1].date, level=1, inplace=True)
                the_dict = self.dts.fund_indicator_annual.to_dict(orient='index')
                _res = self.dts.fund_indicator[['fund_id', 'datetime']].apply(lambda x : self._make_annual_fac(the_dict, x), axis=1)
                _tm_process_fund_annual_apply = time.time()
                _annual_fac_df = pd.DataFrame(_res.tolist())
                self.dts.fund_indicator = pd.merge(self.dts.fund_indicator,_annual_fac_df,how='left',left_on=['fund_id','datetime'],right_on=['fund_id','datetime'])
                _tm_process_fund_annual_join = time.time()
            else:
                _tm_process_indicator_weekly = time.time()
                _tm_process_indicator_monthly = _tm_process_indicator_weekly
                _tm_process_indicator_join = _tm_process_indicator_weekly
                _tm_process_fund_annual_apply = _tm_process_indicator_weekly
                _tm_process_fund_annual_join = _tm_process_indicator_weekly

            self.dts._barra_cne5_factor_return = fetch_table(derived_session, BarraCNE5FactorReturn)
            _tm_deriv_fetch_barra_cne5_factor_return = time.time()

            if print_time:
                print(f'\t[time][deriv] fetch fund indicator: {_tm_deriv_fetch_fund_indicator - _tm_deriv_start}')
                print(f'\t[time][deriv] apply fund indicator: {_tm_deriv_apply_fund_indicator - _tm_deriv_fetch_fund_indicator}')
                print(f'\t[time][deriv] fetch fund indicator annual: {_tm_deriv_fetch_fund_indicator_annual - _tm_deriv_apply_fund_indicator}')
                print(f'\t[time][deriv] fetch index val: {_tm_deriv_fetch_index_pct - _tm_deriv_fetch_fund_indicator_annual}')
                print(f'\t[time][deriv] pivot raw index: {_tm_deriv_apply_index_pct - _tm_deriv_fetch_index_pct}')
                print(f'\t[time][deriv] fetch active indicator score:  {_tm_deriv_apply_active_indicator - _tm_deriv_apply_index_pct}')
                print(f'\t[time][deriv] fetch manager info:  {_tm_deriv_apply_manager_info - _tm_deriv_apply_active_indicator}')
                print(f'\t[time][deriv] fetch manager rank :  {_tm_deriv_apply_manager_rank - _tm_deriv_apply_manager_info}')
                print(f'\t[time][deriv] process indicator weekly: {_tm_process_indicator_weekly - _tm_deriv_apply_manager_rank}')
                print(f'\t[time][deriv] process indicator monthly: {_tm_process_indicator_monthly - _tm_process_indicator_weekly}')
                print(f'\t[time][deriv] process indicator join: {_tm_process_indicator_join - _tm_process_indicator_monthly}')
                print(f'\t[time][deriv] process indicator annually apply: {_tm_process_fund_annual_apply - _tm_process_indicator_join}')
                print(f'\t[time][deriv] merge annual fac: {_tm_process_fund_annual_join - _tm_process_fund_annual_apply}')
                print(f'\t[time][deriv] fetch barra cne5 factor return: {_tm_deriv_fetch_barra_cne5_factor_return - _tm_process_fund_annual_join}')
        _tm_derived = time.time()
        self.dts.fix_data()
        _tm_fix = time.time()

        self._is_inited = True
        self._score_manager.set_dts(self.dts)
        if score_pre_calc:
            self._score_manager.pre_calculate(is_filter_c=True, use_weekly_monthly_indicators=use_weekly_monthly_indicators)
            self._score_manager.pre_calculate_manager()

        # 干掉不用的字段以节省内存
        self.dts.remove_fields(block_df_list)

        _tm_finish = time.time()
        print(self.dts)
        if print_time:
            print(f'[time] basic: {_tm_basic - _tm_start}')
            print(f'[time] deriv: {_tm_derived - _tm_basic}')
            print(f'[time] fix_d: {_tm_fix - _tm_derived}')
            print(f'[time] score: {_tm_finish - _tm_fix}')
            print(f'[time] total: {_tm_finish - _tm_start}')
    
    def _make_annual_fac(self, the_dict, x):
        fund_id = x.fund_id
        dt = x.datetime
        dic = {'fund_id': fund_id, 'datetime': dt}
        last_year = datetime.date(dt.year - 1,12,31)
        last_two_year = datetime.date(dt.year - 2,12,31)
        try:
            dic['last_year_alpha'] = the_dict[(fund_id, last_year)]['alpha_annual']
            dic['last_year_beta'] = the_dict[(fund_id, last_year)]['beta_annual']
            dic['last_two_year_alpha'] = the_dict[(fund_id, last_two_year)]['alpha_annual']
            dic['last_two_year_beta'] = the_dict[(fund_id, last_two_year)]['beta_annual']
        except KeyError:
            pass
        return dic

    def _get_annual_alpha_beta(self, _fund_indicator_annual_fund_i, year):
        if year in _fund_indicator_annual_fund_i.index:
            return _fund_indicator_annual_fund_i.loc[year,'alpha_annual'], _fund_indicator_annual_fund_i.loc[year,'beta_annual']
        else:
            return None, None

    def set_score_param(self, score_param: FundScoreParam):
        self._score_manager.set_param(score_param)

    def get_index_pcts(self, dt):
        # jch: only pct within 7 days take effect
        INDEX_PCT_EFFECTIVE_DELAY_DAY_NUM = 7
        res = {}
        for index_id in self.dts.index_pct.columns:
            df = self.dts.index_pct[index_id]
            _filtered = df[(df.index <= dt) & (df.index >= dt - datetime.timedelta(days=INDEX_PCT_EFFECTIVE_DELAY_DAY_NUM))]
            if len(_filtered) > 0:
                res[index_id] = _filtered.iloc[-1]
        return res

    def get_index_pcts_df(self, dt):
        if dt not in self.dts.index_date_list:
            return pd.DataFrame()
        return self.dts.index_pct.loc[dt]

    def get_fund_score(self, dt, index_id, is_filter_c=True, score_func: Optional[ScoreFunc] = None) -> dict:
        return self._score_manager.get_fund_score(dt, index_id, is_filter_c, score_func)

    def get_fund_scores(self, dt, is_filter_c=True, fund_score_funcs: Optional[Dict[str, Union[ScoreFunc, str]]] = None) -> dict:
        return self._score_manager.get_fund_scores(dt, self.dts.index_list, is_filter_c, fund_score_funcs)

    def get_manager_scores(self, dt, score_select):
        manager_score, manager_funds = self._score_manager.get_manager_scores(dt, self.dts.active_fund_info.fund_id, score_select)
        if not manager_funds:
            return {}, {}, {}
        # 保证主动基金不买到相同基金的基金经理， 只保留高分基金经理， 用于fa helper.on_price -> target funds weight
        # 此分数不用于判定换仓 保留有相同最好基金的基金经理分数
        type_i = score_select['active']
        res = []
        _manager_score = manager_score[type_i]
        for mng_id in _manager_score:
            dic = {
                'mng_id':mng_id,
                'score':_manager_score[mng_id],
                'fund_id':manager_funds[mng_id]
            }
            res.append(dic)
        _df = pd.DataFrame(res).sort_values(['fund_id','score'],ascending = (True, False)).drop_duplicates(['fund_id'], keep='first')
        manager_score_cleaned = _df[['mng_id','score']].set_index('mng_id').to_dict()['score']
        return manager_score, manager_funds, {type_i:manager_score_cleaned}

    def get_fund_nav(self, dt):
        df = self.dts.fund_nav
        return df.loc[dt].to_dict()

    def get_fund_unit_nav(self, dt):
        df = self.dts.fund_unit_nav
        return df.loc[dt].to_dict()

    def confirm_fund_nav_exists(self, fund_ids):
        # param "fund_id" is str (id) or list of str (id_list)
        # get fund nav from db
        fund_list = []
        if type(fund_ids) is list:
            fund_list = fund_ids
        elif type(fund_ids) is str:
            fund_list = [fund_ids]
        fund_list_without_nav = []
        for fund_id in fund_list:
            if fund_id not in self.fund_list_with_nav:
                self.fund_list_with_nav.add(fund_id)
                fund_list_without_nav.append(fund_id)
        # print(f'fund_list: {fund_list}')
        # print(f'fund_list_without_nav: {fund_list_without_nav}')
        if fund_list_without_nav:
            with BasicDatabaseConnector().managed_session() as quant_session:
                _fund_nav_query = quant_session.query(
                    FundNav.fund_id,
                    FundNav.adjusted_net_value,
                    FundNav.unit_net_value,
                    FundNav.datetime
                ).filter(
                    FundNav.fund_id.in_(fund_list_without_nav),
                    FundNav.datetime >= self.start_date,
                    FundNav.datetime <= self.end_date,
                )
                _nav_df = pd.read_sql(_fund_nav_query.statement, _fund_nav_query.session.bind)

                # 这里我们制作一个总的pivot_table再取值，会比分别两次pivot_table快一些
                # 如果index-column pair有重复的值，pivot会报错而pivot_table不会，所以这里用pivot也许更好些
                # 但是pivot里边会调用set_index，导致datetime列的dtype从object变为datetime64[ns]
                _temp_fund_nav = _nav_df.pivot_table(index='datetime', columns='fund_id')

                self.dts.fund_nav = self.dts.fund_nav.join(
                    _temp_fund_nav['adjusted_net_value'].dropna(axis=1, how='all').fillna(method='ffill'),
                    how='outer'
                )
                self.dts.fund_unit_nav = self.dts.fund_unit_nav.join(
                    _temp_fund_nav['unit_net_value'].dropna(axis=1, how='all').fillna(method='ffill'),
                    how='outer'
                )

        print(f'[fund_list_without_nav] fund_ids added : {fund_list_without_nav}')
        return fund_list_without_nav

    def get_fund_purchase_fees(self):
        return self.dts.fund_info.set_index('fund_id').purchase_fee.to_dict()

    def get_fund_redeem_fees(self):
        return self.dts.fund_info.set_index('fund_id').redeem_fee.to_dict()

    def get_fund_info(self):
        return self.dts.fund_info

    def get_trading_days(self):
        return self.dts.trading_days.copy()

    def get_index_price(self, dt=None):
        if dt:
            return self.dts.index_price.loc[dt]
        else:
            return self.dts.index_price.copy()

    def get_index_price_data(self, dt):
        _index_price = self.get_index_price(dt)
        return AssetPrice(**_index_price.to_dict())

    def get_cs_index_component(self) -> pd.DataFrame:
        return self.dts.cs_index_component

    def get_fund_indicator_annual(self) -> pd.DataFrame:
        return self.dts.fund_indicator_annual

    def fund_analysis(self, fund_id: str, is_plot: bool = False) -> Tuple[np.ndarray, pd.Series]:
        from ..fund.derived.barra_cne5_factor_processor import BarraCNE5FactorProcessor
        try:
            params, spec_ret, fund_ret, factor_ret_contri = BarraCNE5FactorProcessor.fund_analysis(self.dts._barra_cne5_factor_return, self.dts.fund_nav.loc[:, fund_id])
        except KeyError:
            return
        if is_plot:
            BarraCNE5FactorProcessor.fund_analysis_plot(fund_ret, factor_ret_contri)
        return params, spec_ret

    def white_and_black_list_filter(self, score, score_raw, disproved_list):
        return self._score_manager._white_and_black_list_filter(score, score_raw, disproved_list)

    def find_same_size_type_funds(self, desc_name):
        fund_info = self.dts.fund_info.copy()
        fund_info['trim_name'] = fund_info.desc_name.map(lambda x : re.subn(r'[ABCDEFR]{1,2}(\(人民币\)|\(美元现汇\)|\(美元现钞\)|1|2|3)?$', '', x)[0])
        fund_info = fund_info.set_index(['trim_name','desc_name'])
        index_1 = fund_info.loc[pd.IndexSlice[:, desc_name], :].index.get_level_values(0).values[0]
        return fund_info.loc[index_1]

def test():
    # m = FundDataManager('20190101', '20200101', score_manager=FundScoreManager())
    m = FundDataManager('20190101', '20200101')
    m.init()
    print(m.get_fund_info())
    print(m.get_trading_days())
    print(m.get_fund_score(datetime.date(2019, 1, 3), 'hs300'))


if __name__ == '__main__':
    with profiled(file_name='/tmp/test.txt'):
        m = FundDataManager('20190101', '20200101', score_manager=FundScoreManager())
        m.init()




import datetime
import traceback
from typing import List, Dict

import pandas as pd
import numpy as np
from dateutil.relativedelta import relativedelta

from ...api.raw import RawDataApi
from ...api.basic import BasicDataApi
from ...view.basic_models import BarraCNE5RiskFactor
from .basic_data_helper import BasicDataHelper


class BarraCNE5FactorData:

    _EXTRA_DAYS = 240

    def __init__(self, _data_helper: BasicDataHelper):
        self._data_helper: BasicDataHelper = _data_helper
        self._factors: Dict[str, pd.DataFrame] = {}
        self._raw_api = RawDataApi()

    def init(self, start_date: str, end_date: str):
        self._start_date: str = start_date
        self._end_date: str = end_date

        # 由于有rolling, 所以需要从start_date往前多取一些数据
        trading_day: pd.DataFrame = BasicDataApi().get_trading_day_list().set_index('datetime')
        start_date_loc = trading_day.index.get_loc(pd.to_datetime(self._start_date, infer_datetime_format=True).date(), method='ffill')
        assert isinstance(start_date_loc, int), 'start date is not a valid trading day'

        assert start_date_loc >= self._EXTRA_DAYS, 'there is no enough extra days before start date'
        start_date_with_extra_days = trading_day.index.array[start_date_loc - self._EXTRA_DAYS]
        print(f'start date with extra days: {start_date_with_extra_days}')
        self._last_trading_day = pd.to_datetime(trading_day.index.array[start_date_loc - 1])

        # 获取未复权收盘价
        stock_price: pd.DataFrame = self._raw_api.get_em_stock_price(start_date_with_extra_days, self._end_date, columns=['close', 'turn']).set_index(['stock_id', 'datetime'])
        self._stock_price: pd.DataFrame = stock_price.drop(columns='turn')
        # 干掉当天没有进行交易的股票, 但这样的话往前取的数据可能不够rolling窗口使用了
        # self._stock_price[~self._stock_price['trade_status'].isin(('停牌一天', '暂停上市', '连续停牌'))]
        print('load stock price done')

        # 获取后复权收盘价及前收
        self._stock_post_price = self._raw_api.get_em_stock_post_price(start_date_with_extra_days, self._end_date, columns=['close', 'pre_close']).set_index(['stock_id', 'datetime'])
        # 按道理索引应该与未复权df完全一致，但可能值是一样的但顺序不同，这里reindex一下
        self._stock_post_price = self._stock_post_price.reindex(index=self._stock_price.index)
        # 干掉当天没有进行交易的股票, 但这样的话往前取的数据可能不够rolling窗口使用了
        # self._stock_post_price[~self._stock_post_price['trade_status'].isin(('停牌一天', '暂停上市', '连续停牌'))]
        print('load stock post price done')

        pd.testing.assert_index_equal(self._stock_post_price.index, self._stock_price.index)
        # reindex可能掩盖了不一致，这里看一下是否有某一行除了索引之外全为null
        assert not self._stock_post_price.isnull().all(axis=1).any(), 'stock post price df should not have any NaN!!'

        # 获取换手率
        self._stock_turnover: pd.DataFrame = stock_price.drop(columns='close')
        print('load stock turnover done')

        # 获取每日更新的总股本
        self._stock_daily_info: pd.DataFrame = self._raw_api.get_em_daily_info(start_date_with_extra_days, self._end_date, columns=['total_share']).set_index(['stock_id', 'datetime'])
        # 与未复权股价的索引对齐(空值保持为nan)
        self._stock_daily_info = self._stock_daily_info.reindex(index=self._stock_price.index)
        # 对齐后将每只个股时间序列上总股本为nan的值ffill, TODO: 再做bfill了，这样对吗
        self._stock_daily_info['total_share'] = self._stock_daily_info.reset_index().pivot_table(index='stock_id', columns='datetime', values='total_share', dropna=False).fillna(method='ffill', axis=1).fillna(method='bfill', axis=1).stack()
        print('load stock daily info done')
        pd.testing.assert_index_equal(self._stock_daily_info.index, self._stock_price.index)
        assert not self._stock_daily_info.isnull().to_numpy().any(), 'stock daily info df should not have any NaN!!'

        # 获取个股财务数据
        equity_start: datetime.date() = (pd.to_datetime(self._start_date, infer_datetime_format=True) - relativedelta(years=5)).date()
        self._fin_fac = self._raw_api.get_em_stock_fin_fac(start_date=equity_start, end_date=self._end_date)
        print('load fin_fac info done')

    # 5%-95%
    @staticmethod
    def winsorize_quantile(x: pd.DataFrame, flag=0.05) -> pd.DataFrame:
        x = x.stack()
        x = x[x.notna().all(axis=1)]
        if x.empty:
            return
        high_low_quantile = x.quantile([1 - flag, flag])
        anew = x.droplevel(0)
        anew = anew.where(anew <= high_low_quantile.loc[1 - flag, :], high_low_quantile.loc[1 - flag, :], axis=1)
        anew = anew.where(anew >= high_low_quantile.loc[flag, :], high_low_quantile.loc[flag, :], axis=1)
        return anew

    def calculate_factors(self):
        # 日对数收益率
        temp_return: pd.Series = np.log(self._stock_post_price.loc[:, 'close'] / self._stock_post_price.loc[:, 'pre_close'])
        self._factors['ret'] = temp_return.unstack(level=0)
        self._factors['momentum'] = self._factors['ret'].rolling(20).sum()

        # 市值因子：size取对数
        temp_shareweight: pd.DataFrame = (self._stock_daily_info.loc[:, 'total_share'] * self._stock_price.loc[:, 'close']).unstack(level=0)
        self._factors['logsize'] = -np.log(temp_shareweight)

        fin_fac_pivoted_df: pd.DataFrame = self._fin_fac.pivot_table(index='datetime', columns='stock_id', values=['balance_statement_140', 'income_statement_61', 'income_statement_83', 'cashflow_statement_39', 'balance_statement_74', 'balance_statement_128'], dropna=False)
        total_equity: pd.DataFrame = fin_fac_pivoted_df['balance_statement_140']
        # fin_fac(财务数据)有一部分数据的日期可能是非交易日，这里取并fill一下，然后再回到交易日的index上来
        total_equity = total_equity.reindex(index=self._factors['ret'].index.union(total_equity.index)).fillna(method='ffill').reindex(index=self._factors['ret'].index)

        # 计算有效股票列表
        effective_col: pd.Index = temp_shareweight.columns.intersection(total_equity.columns)
        self._factors['ret'] = self._factors['ret'].loc[:, effective_col]
        temp_shareweight = temp_shareweight.loc[:, effective_col]

        self._factors['book_price'] = total_equity.loc[:, effective_col] / temp_shareweight

        # index: datetime, value: float
        temp_rm: pd.Series = (self._factors['ret'] * temp_shareweight / temp_shareweight.sum()).sum(axis=1)
        beta = self._factors['ret'].rolling(20).cov(temp_rm).div(temp_rm.rolling(20).var(), axis=0)
        DASTD = self._factors['ret'].rolling(20).std()
        CMRA = self._factors['ret'].rolling(20).max() - self._factors['ret'].rolling(20).min()
        HSIGMA = (self._factors['ret'] - beta.mul(temp_rm, axis=0)).rolling(20).std()
        self._factors['resvol'] = 0.74 * DASTD + 0.16 * CMRA + 0.1 * HSIGMA

        net_earnings: pd.DataFrame = fin_fac_pivoted_df['income_statement_61'].loc[:, effective_col]
        EGRO = net_earnings.rolling(5).apply(lambda x: (x[-1] - x[0]) / x[0], raw=True)
        net_sales: pd.DataFrame = fin_fac_pivoted_df['income_statement_83'].loc[:, effective_col]
        SGRO = net_sales.rolling(5).apply(lambda x: (x[-1] - x[0]) / x[0], raw=True)
        growth = 0.34 * EGRO + 0.66 * SGRO
        self._factors['growth'] = growth.reindex(index=self._factors['resvol'].index.union(growth.index)).fillna(method='ffill').reindex(index=self._factors['resvol'].index)

        temp_turnover = self._stock_turnover.pivot_table(index='datetime', columns='stock_id', values='turn', dropna=False).fillna(method='ffill')
        STOM = temp_turnover.rolling(20).sum()
        # 这里rolling.apply很慢，可以考虑用注释中的近似算法，不过目前engine用numba且每日更新数据量较小，耗时可以接受
        STOQ = temp_turnover.rolling(60).apply(lambda x: (x * (1 / np.arange(60//20, 0, -1)).repeat(20)).sum(), engine='numba', raw=True)
        # STOQ = temp_turnover.rolling(60).sum() / 3
        STOA = temp_turnover.rolling(240).apply(lambda x: (x * (1 / np.arange(240//20, 0, -1)).repeat(20)).sum(), engine='numba', raw=True)
        # STOA = temp_turnover.rolling(240).sum() / 12
        self._factors['liquidity'] = 0.35 * np.log(STOM) + 0.35 * np.log(STOQ) / 3 + 0.3 * np.log(STOA) / 12

        ETOP: pd.DataFrame = net_earnings.reindex(index=self._factors['resvol'].index.union(net_earnings.index)).fillna(method='ffill').div(temp_shareweight).reindex(index=self._factors['resvol'].index)
        net_cash: pd.DataFrame = fin_fac_pivoted_df['cashflow_statement_39']
        CETOP: pd.DataFrame = net_cash.reindex(index=self._factors['resvol'].index.union(net_cash.index)).fillna(method='ffill').div(temp_shareweight).reindex(index=self._factors['resvol'].index)
        self._factors['earning_yields'] = 0.66 * ETOP + 0.34 * CETOP

        leverage: pd.DataFrame = fin_fac_pivoted_df['balance_statement_128'] / fin_fac_pivoted_df['balance_statement_74']
        self._factors['leverage'] = leverage.reindex(self._factors['resvol'].index.union(leverage.index)).fillna(method='ffill').reindex(self._factors['resvol'].index)
        print(f'all factors are calculated successfully')

        # 将所有因子放到一个df里
        concat_keys: List[str] = []
        df_list: List[pd.Series] = []
        for name, factor in self._factors.items():
            # ret这里先不放进来
            if name != 'ret':
                factor = factor.loc[:, effective_col]
                factor = factor[factor.index >= self._last_trading_day]
                df_list.append(factor)
                concat_keys.append(name)
        whole_df: pd.DataFrame = pd.concat(df_list, axis=1, keys=concat_keys)
        whole_df = whole_df.shift(1).groupby(by='datetime', sort=False).apply(BarraCNE5FactorData.winsorize_quantile, flag=0.01)
        ret = self._factors['ret']
        ret = ret[ret.index >= self._last_trading_day]
        self._whole_factors = whole_df.join(ret.stack().rename('ret'), how='outer')

    def process_all(self, start_date: str, end_date: str):
        start_date = pd.to_datetime(start_date, infer_datetime_format=True).date().isoformat()
        end_date = pd.to_datetime(end_date, infer_datetime_format=True).date().isoformat()
        failed_tasks: List[str] = []
        try:
            self.init(start_date, end_date)
            self.calculate_factors()
            self._data_helper._upload_basic(self._whole_factors.loc[start_date:end_date, :].reset_index(), BarraCNE5RiskFactor.__table__.name)
        except Exception as e:
            print(e)
            traceback.print_exc()
            failed_tasks.append(BarraCNE5RiskFactor.__table__.name)
        return failed_tasks




from ..structure import Tick
from .wrapper import DataWrapper
import pandas as pd
import numpy as np
import os
import json
import datetime 

class RQDataWrapper(DataWrapper):

    TICK_COLUMNS = [
        'open', 'last', 'high', 'low', 'prev_close', 'volume', 'total_turnover',
        'limit_up', 'limit_down', 'a1', 'a2', 'a3', 'a4', 'a5', 'b1', 'b2',
        'b3', 'b4', 'b5', 'a1_v', 'a2_v', 'a3_v', 'a4_v', 'a5_v', 'b1_v',
        'b2_v', 'b3_v', 'b4_v', 'b5_v', 'change_rate', 'time'
    ]

    def __init__(self, base_path):
        super().__init__(base_path)

    def load_data(self, month, ticker):  
        folder = os.path.join(self.base_path, month)
        return DataWrapper.base_load_data(folder, ticker, 'time', self.TICK_COLUMNS)

    def parse(self, index, row):
        return Tick(ticker=row.ticker, mkt_time= pd.to_datetime(index, format='%Y-%m-%dT%H:%M:%S.%fZ'),
                    bid_price1=row.b1, ask_price1=row.a1, bid_volume1=row.b1_v, ask_volume1=row.a1_v,
                    bid_price2=row.b2, ask_price2=row.a2, bid_volume2=row.b2_v, ask_volume2=row.a2_v,
                    bid_price3=row.b3, ask_price3=row.a3, bid_volume3=row.b3_v, ask_volume3=row.a3_v,
                    bid_price4=row.b4, ask_price4=row.a4, bid_volume4=row.b4_v, ask_volume4=row.a4_v,
                    bid_price5=row.b5, ask_price5=row.a5, bid_volume5=row.b5_v, ask_volume5=row.a5_v,
                    up_limit=row.limit_up, down_limit=row.limit_down, 
                    last_price=row['last']
        )
    

class OTTicksWrapper(DataWrapper):

    TICK_COLUMNS = [
        'ask_price_1', 'ask_volume_1', 'ask_price_2', 'ask_volume_2', 'ask_price_3', 'ask_volume_3',
        'ask_price_4', 'ask_volume_4', 'ask_price_5', 'ask_volume_5', 'bid_price_1', 'bid_volume_1',
        'bid_price_2', 'bid_volume_2', 'bid_price_3', 'bid_volume_3', 'bid_price_4', 'bid_volume_4',
        'bid_price_5', 'bid_volume_5', 'last_price',  'mkt_time', 'volume',
    ]

    def __init__(self, base_path):
        super().__init__(base_path)

    def load_data(self, month, ticker): 
        folder = os.path.join(self.base_path, month)
        df = DataWrapper.base_load_data(folder, ticker, 'mkt_time', self.TICK_COLUMNS)
        df.index = pd.to_datetime(df.index, unit='ns')
        return df

    def parse(self, index, row):
        return Tick(ticker=row.ticker, mkt_time= pd.to_datetime(index, format='%Y-%m-%dT%H:%M:%S.%fZ'),
                    last_price=row.last_price,
                    ask_price1=row.ask_price_1, ask_volume1=row.ask_volume_1, bid_price1=row.bid_price_1, bid_volume1=row.bid_volume_1,
                    ask_price2=row.ask_price_2, ask_volume2=row.ask_volume_2, bid_price2=row.bid_price_2, bid_volume2=row.bid_volume_2,
                    ask_price3=row.ask_price_3, ask_volume3=row.ask_volume_3, bid_price3=row.bid_price_3, bid_volume3=row.bid_volume_3,
                    ask_price4=row.ask_price_4, ask_volume4=row.ask_volume_4, bid_price4=row.bid_price_4, bid_volume4=row.bid_volume_4,
                    ask_price5=row.ask_price_5, ask_volume5=row.ask_volume_5, bid_price5=row.bid_price_5, bid_volume5=row.bid_volume_5
        )


class StockDailyWrapper():

    def __init__(self,base_path=None,price_type=None,stock_pool=None,start_time=None,end_time=None,factor_path=None,factor_name=None,vwap_limit=None):
        self.base_path       = base_path
        self.price_type      = price_type
        self.stock_pool      = stock_pool
        self.factor_path     = factor_path
        self.factor_name     = factor_name
        self.vwap_limit      = vwap_limit
        self.start_time      = self.change_date_time(start_time)
        self.end_time        = self.change_date_time(end_time)
        self.stock_pool_data = self.load_stock_pool()
        self.common_pool     = self.load_common_pool()
        self.price_data      = self.load_price()
        self.st_data         = self.load_st()
        self.suspended_data  = self.load_suspended()
        self.stock_info      = self.load_stock_info()
        self.date_list       = sorted(list(self.price_data.keys())) 
        self.factor_data     = self.load_factor()
        self.vwap_diff_limit_stock = self.load_vwap_diff_limit_stock()

    def change_date_time(self, str_date):
        if len(str_date) == 8:
            return str_date[:4]+'-'+str_date[4:6]+'-'+str_date[6:8]
        else:
            return str_date

    def load_json(self, paths):
        with open(paths,'r') as load_f:
            json_data = json.load(load_f)
        return json_data
        
    def load_stock_pool(self):
        paths = os.path.join(self.base_path, 'stock_pool', f'{self.stock_pool}.json')
        dic = self.load_json(paths)
        return self.filter_dict_by_key(dic)

    def load_common_pool(self):
        common_pool = set()
        for i in list(self.stock_pool_data.values()):
            common_pool.update(i)
        return list(common_pool)

    def load_stock_info(self):
        paths = os.path.join(self.base_path, 'stock_info', 'stock_info.json')
        return self.load_json(paths)

    def load_price(self):
        paths = os.path.join(self.base_path, self.price_type)
        meta_info = DataWrapper(paths).meta_info
        columns = meta_info['tickers']+['time']
        df = DataWrapper.base_load_npz(paths, self.price_type, columns)
        df = df.fillna(method='ffill')
        df.index = df.time.dt.strftime('%Y-%m-%d')
        df.drop(columns=['time'],inplace=True)
        df = df[self.common_pool]
        df = df.astype(float).round(3).copy()
        df['time'] = df.index
        data_dic = df.to_dict('index')
        return self.filter_dict_by_key(data_dic)

    def load_st(self):
        paths = os.path.join(self.base_path, 'st', 'st.json')
        dic = self.load_json(paths)
        return self.filter_dict_by_key(dic)

    def load_suspended(self):
        paths = os.path.join(self.base_path, 'suspended', 'suspended.json')
        dic = self.load_json(paths)
        return self.filter_dict_by_key(dic)
        
    def load_factor(self):
        factor_path = os.path.join(self.factor_path, f'{self.factor_name}.npz')
        factor_meta_path = os.path.join(self.factor_path, '.meta_data.json')
        meta_info = self.load_json(factor_meta_path)
        value = np.load(factor_path, allow_pickle=True)
        df = pd.DataFrame(data=value['arr_0'], columns=meta_info['tickers']+['time'])
        df.index = df.time.dt.strftime('%Y-%m-%d')
        df.drop(columns=['time'],inplace=True)
        try:
            df = df[self.common_pool]
        except:
            pass
        df = df.astype(float).round(4).copy()
        df['time'] = df.index
        data_dic = df.to_dict('index')
        return self.filter_dict_by_key(data_dic)

    def load_black_list(self):
        pass

    def load_vwap_diff_limit_stock(self):
        paths = os.path.join(self.base_path, 'post_vwap_preclose_diff')
        meta_info = DataWrapper(paths).meta_info
        columns = meta_info['tickers']+['time']
        df = DataWrapper.base_load_npz(paths, 'post_vwap_preclose_diff', columns)
        df.index = df.time.dt.strftime('%Y-%m-%d')
        df.drop(columns=['time'],inplace=True)
        df = df[self.common_pool]
        df = df.astype(float).round(2).abs().copy()
        df['time'] = df.index
        data_dic = df.to_dict('index')
        return self.filter_dict_by_key(data_dic)

    def remove_limit_vwap_stock(self, date):
        dic = self.vwap_diff_limit_stock[date]
        del dic['time']
        rm_l = [k for k, v in  dic.items() if v > self.vwap_limit]
        return rm_l

    def filter_dict_by_key(self, dic):
        dic = dict(filter(lambda elem: (elem[0] >= self.start_time and elem[0] <= self.end_time), dic.items()))
        return dic

    def find_beginday_yesterday(self):
        paths = os.path.join(self.base_path, self.price_type)
        meta_info = DataWrapper(paths).meta_info
        start_time = [ _ for _ in meta_info['dates'] if self.start_time < _][0]
        d = meta_info['dates'][meta_info['dates'].index(start_time) - 1]       
        return d
    


from .constant import OrderStatus, ExecRole, ExchangeID, Direction, OrderType, AssetType, PosiDirection
from collections import deque
import numpy as np
import yaml
import dataclasses

class TickerInfo:

    def __init__(self):
        self.loc = 0
        self.file = ' '
        self.file_reader_time = 0
        self.df = ' '
        self.len = 0
        self.file_end_time = 0

class Sell():
    def __init__(self, ticker, volume):
        self.ticker = ticker
        self.volume = volume

class Tick:

    def __init__(self, **kwds):
        self.ticker = kwds.get('ticker')
        self.up_limit = kwds.get('up_limit')
        self.down_limit = kwds.get('down_limit')
        self.last_price = kwds.get('last_price')
        self.mkt_time = kwds.get('mkt_time') # datetime
        self.bid_price1 = kwds.get('bid_price1')
        self.ask_price1 = kwds.get('ask_price1')
        self.bid_volume1 = kwds.get('bid_volume1')
        self.ask_volume1 = kwds.get('ask_volume1')
        self.bid_price2 = kwds.get('bid_price2')
        self.ask_price2 = kwds.get('ask_price2')
        self.bid_volume2 = kwds.get('bid_volume2')
        self.ask_volume2 = kwds.get('ask_volume2')
        self.bid_price3 = kwds.get('bid_price3')
        self.ask_price3 = kwds.get('ask_price3')
        self.bid_volume3 = kwds.get('bid_volume3')
        self.ask_volume3 = kwds.get('ask_volume3')
        self.bid_price4 = kwds.get('bid_price4')
        self.ask_price4 = kwds.get('ask_price4')
        self.bid_volume4 = kwds.get('bid_volume4')
        self.ask_volume4 = kwds.get('ask_volume4')
        self.bid_price5 = kwds.get('bid_price5')
        self.ask_price5 = kwds.get('ask_price5')
        self.bid_volume5 = kwds.get('bid_volume5')
        self.ask_volume5 = kwds.get('ask_volume5')

    def __str__(self):
        return '<Tick ticker={} tm={} b1={} a1={} b1v={} a1v={} last_price={}>'.format(
            self.ticker, self.mkt_time.strftime('%Y%m%d-%H:%M:%S.%f'),
            self.bid_price1, self.ask_price1, self.bid_volume1, self.ask_volume1, self.last_price
        )

class Order:

    def __init__(self, **kwds):
        self.order_id = kwds.get('order_id')
        self.exchange = kwds.get('exchange', ExchangeID.NOT_AVAILABLE)
        self.ticker = kwds.get('ticker')
        self.asset_type = kwds.get('asset_type', AssetType.NOT_AVAILABLE)
        self.direction = kwds.get('direction', Direction.NOT_AVAILABLE)
        self.order_type = kwds.get('order_type', OrderType.NOT_AVAILABLE)
        self.price = kwds.get('price') # price is None means order_type has to be MARKET
        self.volume = kwds.get('volume')
        self.source = kwds.get('source')
        self.err_msg = kwds.get('err_msg')
        self.status = kwds.get('status', OrderStatus.NOT_AVAILABLE)
        
    def __str__(self):
        return '<Order id={} ticker={} direction={} ot={} at={} ex={} price={} volume={} source={} err_msg={}>'.format(
            self.order_id, self.ticker, Direction.read(self.direction), OrderType.read(self.order_type), 
            AssetType.read(self.asset_type), ExchangeID.read(self.exchange), self.price, self.volume, self.source, self.err_msg
        )


class Trade:

    def __init__(self, **kwds):
        self.order_id = kwds.get('order_id')
        self.exchange = kwds.get('exchange', ExchangeID.NOT_AVAILABLE)
        self.ticker = kwds.get('ticker')
        self.asset_type = kwds.get('asset_type', AssetType.NOT_AVAILABLE)
        self.direction = kwds.get('direction', Direction.NOT_AVAILABLE)
        self.trade_price = kwds.get('trade_price')
        self.trade_volume = kwds.get('trade_volume')
        self.exec_role = kwds.get('exec_role', ExecRole.NOT_AVAILABLE)
        self.trade_id = kwds.get('trade_id')
        self.source = kwds.get('source')
        self.time = kwds.get('time','')

    def __str__(self):
        return f'<Trade id={self.trade_id} exchange={ExchangeID.read(self.exchange)} ticker={self.ticker} direction={Direction.read(self.direction)} at={AssetType.read(self.asset_type)} ' \
               f'price={self.trade_price} volume={self.trade_volume} source={self.source} role={ExecRole.read(self.exec_role)} order_id={self.order_id} time={self.time}>'

    def save_trade(self):
        return {'id'        :   self.trade_id, 
                'exchange'  :   ExchangeID.read(self.exchange),
                'ticker'    :   self.ticker,
                'direction' :   self.direction,
                'asset_type':   self.asset_type,
                'price'     :   self.trade_price,
                'volume'    :   self.trade_volume,
                'time'      :   self.time}
                
class BacktestKey:

    def __init__(self, **kwds):
        self.str_name   = kwds.get('strategy_name')
        self.factor     = kwds.get('factor')
        self.pool       = kwds.get('pool')
        self.longshort  = kwds.get('longshort')
        self.top        = kwds.get('top')
        self.period     = kwds.get('period')
        self.buy_price  = kwds.get('buy_price')
        self.sell_price = kwds.get('sell_price')
        self.fee        = kwds.get('fee')
        self.start      = kwds.get('start')
        self.end        = kwds.get('end')
        self.vwap_limit = kwds.get('vwap_limit')

    def key(self):
        return f'{self.str_name} {self.factor} {self.pool} {self.longshort} {self.top} {self.period} {self.buy_price} ' \
               f'{self.sell_price} {self.fee} {self.start} {self.end} {self.vwap_limit}'





class TransferRequest:

    def __init__(self, **kwds):
        self.transfer_id = kwds.get('transfer_id')
        self.from_ticker = kwds.get('from_ticker')
        self.to_ticker = kwds.get('to_ticker')
        self.asset_type = kwds.get('asset_type', AssetType.NOT_AVAILABLE)
        self.from_exchange = kwds.get('from_exchange')
        self.to_exchange = kwds.get('to_exchange')
        self.volume = kwds.get('volume')
        self.source = kwds.get('source')
        self.err_msg = kwds.get('err_msg')

    def __str__(self):
        return f'<Transfer id={self.transfer_id} from_ticker={self.from_ticker}  to_ticker={self.to_ticker} at={AssetType.read(self.asset_type)} from={self.from_exchange} ' \
               f'to={self.to_exchange} volume={self.volume} source={self.source}>'


class PosItem:

    def __init__(self, **kwds):
        self.exchange = kwds.get('exchange', ExchangeID.NOT_AVAILABLE)
        self.ticker = kwds.get('ticker')
        self.asset_type = kwds.get('asset_type', AssetType.NOT_AVAILABLE)
        self.posi_direction = kwds.get('posi_direction', PosiDirection.NET)
        self.frozen = kwds.get('frozen', 0)
        self.position = kwds.get('position', 0)
        self.yd_position = kwds.get('yd_position', 0)
        self.to_transfer = kwds.get('to_transfer', 0)
        self.fee_rate_type = kwds.get('fee_rate_type')
        self.avg_cost = kwds.get('avg_cost', 0)
        # 以下两项只有当asset_type是FUTURE_MARGIN时候才有
        self.realized_margin = kwds.get('realized_margin', 0)
        self.unrealized_margin = kwds.get('unrealized_margin', 0)
        self.time = kwds.get('time','')
        # 如果这是期货合约的话,还需要标注不同保证金,因为即使同样是BTC.USD合约,用BTC结算和USDT结算都应该算作不同的合约
        # 如果这posItem是保证金的话,为了计算它的Unrealized_margin,需要把最近一段时间的价格都存进来算平均(防止插针爆仓)
        # 储存这个价格序列好像也就只有在这作为这个合约posItem的attribute比较合适了
        self.margin = kwds.get('margin')
        self.cal_unrealized_len = 0
        self.clear_prices = None
        self.mean_price = 0
        self.leverage = 0

    def get_price_ticker(self):
        # todo: based on different asset type, 
        #   we should use different price
        return self.ticker

    def get_market_value(self, last_price=None):
        if self.asset_type == AssetType.CRYPTO_CONTRACT_MARGIN:
            return self.realized_margin + self.unrealized_margin
        else:
            return self.position * last_price * (-1 if self.posi_direction is PosiDirection.SHORT else 1)

    def transfer(self, volume):
        if self.asset_type == AssetType.CRYPTO_CONTRACT_MARGIN:
            if self.realized_margin + self.unrealized_margin - self.frozen - self.to_transfer >= volume:
                self.realized_margin -= volume
                self.to_transfer += volume
                return True
        if self.position - self.frozen - self.to_transfer >= volume:
            self.position -= volume
            self.to_transfer += volume
            return True
        else:
            return False

    def switch_day(self):
        self.yd_position = self.position
   
    def init_margin(self, pi):
        # 如果当前item是个contract,在此设定这个contract对应的margin
        self.margin = pi
        self.cal_unrealized_len = 100
        self.clear_prices = deque(maxlen=self.cal_unrealized_len)
        self.mean_price = 0

    def append_clear(self, tick):
        self.clear_prices.append((tick.ask_price1 + tick.bid_price1) / 2)
        self.mean_price = np.mean(self.clear_prices)
        if self.position != 0:
            self.margin.unrealized_margin = self.position * PositionInfo.CONTRACT_SIZES[ExchangeID.read(self.exchange)] * (1 / self.avg_cost - 1 / self.mean_price)
            self.leverage = abs(self.position * PositionInfo.CONTRACT_SIZES[ExchangeID.read(self.exchange)] / self.mean_price) / (self.margin.realized_margin + self.margin.unrealized_margin)

    def __str__(self):
        return '<PosItem ex={} ticker={} at={} d={} pos={} yd={} trans={} margin={} time={}>'.format(
            ExchangeID.read(self.exchange), self.ticker, AssetType.read(self.asset_type), 
            PosiDirection.read(self.posi_direction), self.position, self.yd_position, self.to_transfer, self.margin, self.time
        )

class PositionInfo:
    
    path1 = '../python/config'
    path2 = './config'
    try:
        with open(f'{path1}/fee_rate.yml', 'r') as f:
            FEE_RATE = yaml.safe_load(f)
        with open(f'{path1}/contract_sizes.yml', 'r') as f:
            CONTRACT_SIZES = yaml.safe_load(f)
    except:
        with open(f'{path2}/fee_rate.yml', 'r') as f:
            FEE_RATE = yaml.safe_load(f)
        with open(f'{path2}/contract_sizes.yml', 'r') as f:
            CONTRACT_SIZES = yaml.safe_load(f)


    def __init__(self, **kwds):
        # 不同交易所不同同币种手续费率都不一样,因此在这更合适的方法是将手续费率用config形式读入,需要新的交易币种交易方式在此添加即可
        
        self.cash = kwds.get('cash', PosItem())
        _holdings = kwds.get('holdings', [])
        self.holdings = []
        self._tick_idx = {}
        self.stock_buy_fee  =  self.FEE_RATE['STOCK']['BUYFEE']
        self.stock_sell_fee =  self.FEE_RATE['STOCK']['SELLFEE']
        for pi in _holdings:
            self.create_holding(pi)
        # 把合约与对应保证金的关系在这里明确好
        for pi in _holdings:
            if pi.asset_type == AssetType.CRYPTO_COIN_MARGIN_CONTRACT:
                pi.init_margin(self.get_holding(ticker=pi.margin, exchange=pi.exchange))

    def get_holding(self, ticker, exchange=ExchangeID.NOT_AVAILABLE):
        idx = self._tick_idx.get((ticker, exchange))
        return None if idx is None else self.holdings[idx] 

    def create_holding(self, pos_item):
        self.holdings.append(pos_item)
        self._tick_idx[(pos_item.ticker, pos_item.exchange)] = len(self.holdings) - 1

    @staticmethod
    def get_commission(trade, pos):
        rate = PositionInfo.FEE_RATE[ExchangeID.read(trade.exchange)][AssetType.read(trade.asset_type)][pos.fee_rate_type][-1][ExecRole.read(trade.exec_role)]
        commission = rate * trade.trade_volume * PositionInfo.CONTRACT_SIZES[ExchangeID.read(trade.exchange)] / trade.trade_price
        return commission

    def process_trade(self, trade):

        if trade.asset_type >= AssetType.CRYPTO_ASSET:
            # 只处理倒数合约情况
            pos = self.get_holding(ticker=trade.ticker, exchange=trade.exchange)
            margin_pos = pos.margin
            if pos is None:
                pos = PosItem(exchange=trade.exchange, ticker=trade.ticker, asset_type=trade.asset_type)
                self.create_holding(pos)
            if pos.position < 0:
                if trade.direction == Direction.SELL:
                    new_position = pos.position - trade.trade_volume
                    pos.avg_cost = new_position / (pos.position / pos.avg_cost - trade.trade_volume / trade.trade_price)
                    pos.position = new_position
                    margin_pos.realized_margin -= self.get_commission(trade, pos)
                elif trade.direction == Direction.BUY:
                    if trade.trade_volume + pos.position > 0:
                        margin_pos.realized_margin -= self.get_commission(trade, pos)
                        margin_pos.realized_margin += pos.position * PositionInfo.CONTRACT_SIZES[ExchangeID.read(trade.exchange)] * ((1 / pos.avg_cost) - (1 / trade.trade_price))
                        pos.position += trade.trade_volume
                        pos.avg_cost = trade.trade_price
                    else:
                        pos.position += trade.trade_volume
                        margin_pos.realized_margin += trade.trade_volume * PositionInfo.CONTRACT_SIZES[ExchangeID.read(trade.exchange)] * ((1 / trade.trade_price) - (1 / pos.avg_cost)) \
                            - self.get_commission(trade, pos)
            elif pos.position > 0:
                if trade.direction == Direction.BUY:
                    new_position = pos.position + trade.trade_volume
                    pos.avg_cost = new_position / (pos.position / pos.avg_cost + trade.trade_volume / trade.trade_price)
                    pos.position = new_position
                    margin_pos.realized_margin -= self.get_commission(trade, pos)
                elif trade.direction == Direction.SELL:
                    if pos.position - trade.trade_volume < 0:
                        margin_pos.realized_margin -= self.get_commission(trade, pos)
                        margin_pos.realized_margin += pos.position * PositionInfo.CONTRACT_SIZES[ExchangeID.read(trade.exchange)] * ((1 / pos.avg_cost) - (1 / trade.trade_price))
                        pos.position -= trade.trade_volume
                        pos.avg_cost = trade.trade_price
                    else:
                        pos.position -= trade.trade_volume
                        margin_pos.realized_margin += trade.trade_volume * PositionInfo.CONTRACT_SIZES[ExchangeID.read(trade.exchange)] * ((1 / pos.avg_cost) - (1 / trade.trade_price)) - \
                            self.get_commission(trade, pos)
            else:
                margin_pos.realized_margin -= self.get_commission(trade, pos)
                pos.avg_cost = trade.trade_price
                if trade.direction == Direction.BUY:
                    pos.position = trade.trade_volume
                elif trade.direction == Direction.SELL:
                    pos.position = -trade.trade_volume
            return
        else:
            pos = self.get_holding(ticker=trade.ticker, exchange=trade.exchange)    
            if pos is None:
                pos = PosItem(exchange=trade.exchange, ticker=trade.ticker,
                            asset_type=trade.asset_type)
                self.create_holding(pos)
            pos.position += trade.trade_volume if trade.direction == Direction.BUY else - trade.trade_volume
            if pos.asset_type in [AssetType.STOCK]:
                trade_amount = trade.trade_volume * trade.trade_price
                if trade.direction == Direction.BUY:
                    trade_amount = -trade_amount * ( 1 + self.stock_buy_fee)
                else:
                    trade_amount = trade_amount * (1 - self.stock_sell_fee)
                self.cash.position += trade_amount
            # todo: as contract or futures, we should make cash into margin or frozen
            # some logic needs to be confirmed

    def init_transfer(self, transfer):
        from_pos = self.get_holding(ticker=transfer.from_ticker, exchange=transfer.from_exchange)  # 就不担心出现position 比要转的 volume 低的情况吗?下面的 from_pos.transfer检查了
        return False if from_pos is None else from_pos.transfer(transfer.volume)

    def confirm_transfer(self, transfer):
        from_pos = self.get_holding(ticker=transfer.from_ticker, exchange=transfer.from_exchange)
        from_pos.to_transfer -= transfer.volume
        to_pos = self.get_holding(ticker=transfer.to_ticker, exchange=transfer.to_exchange)
        if to_pos is None:
            to_pos = PosItem(exchange=transfer.to_exchange, ticker=transfer.to_ticker, asset_type=transfer.asset_type)
            self.create_holding(to_pos)
        to_pos.realized_margin += transfer.volume

    def process_tick(self, tick):
        # 如果有仓位需要在每个tick之后进行处理的都在此处理
        # 例如对于期货合约,每个tick之后要计算即时保证金的变化,就放在这
        for pi in self.holdings:
            if pi.ticker == tick.ticker:
                if pi.asset_type == AssetType.CRYPTO_COIN_MARGIN_CONTRACT:
                    # 对于期货,需要每个tick结束后计算它的unrealized_margin
                    pi.append_clear(tick)

    def get_total_market_value(self):
        total_market_value = 0
        for pi in self.holdings:
            total_market_value += pi.get_market_value()
        return total_market_value

    def __str__(self):
        msg = '<PositionInfo cash={} \n'.format(self.cash)
        for pi in self.holdings:
            msg += ' {}\n'.format(pi)
        msg += '>'
        return msg



from .cas_connection import CassandraConnector
from cassandra import ConsistencyLevel

class CasClient:

    def __init__(self):
        self.session = CassandraConnector().get_session()
        self.conn = CassandraConnector().get_conn()

    def get_columns(self, model, columns):
        columns_format = '{},' * (len(columns) - 1) + '{}'
        columns_format = columns_format.format(*columns)

        cql = "SELECT {} from {}.{}".format(
            columns_format,
            model.__keyspace__,
            model.__table_name__,
        )
        return self.session.execute(cql)

    def get_columns_by_id(self, model, columns, id_column, id_value):
        columns_format = '{},' * (len(columns) - 1) + '{}'
        columns_format = columns_format.format(*columns)

        cql = "SELECT {} from {}.{} WHERE {}=?".format(
            columns_format,
            model.__keyspace__,
            model.__table_name__,
            id_column,
        )
        cql_lookup = self.session.prepare(cql)
        cql_lookup.consistency_level = ConsistencyLevel.ONE

        ret = self.session.execute(cql_lookup, [id_value])
        ret = [i for i in ret]
        if len(ret) < 1:
            return None

        return ret[0]

    def scan_columns_by_id(self, model, columns, id_column, id_value):
        columns_format = '{},' * (len(columns) - 1) + '{}'
        columns_format = columns_format.format(*columns)

        cql = "SELECT {} from {}.{} WHERE {}=?".format(
            columns_format,
            model.__keyspace__,
            model.__table_name__,
            id_column,
        )
        cql_lookup = self.session.prepare(cql)
        cql_lookup.consistency_level = ConsistencyLevel.ONE

        return [i for i in self.session.execute(cql_lookup, [id_value])]


    def get_columns_by_two_ids(self, model, columns, id_column, id_value, id_column2, id_value2):
        columns_format = '{},' * (len(columns) - 1) + '{}'
        columns_format = columns_format.format(*columns)

        cql = "SELECT {} from {}.{} WHERE {}=? AND {}=?".format(
            columns_format,
            model.__keyspace__,
            model.__table_name__,
            id_column,
            id_column2,
        )
        cql_lookup = self.session.prepare(cql)
        cql_lookup.consistency_level = ConsistencyLevel.ONE

        ret = self.session.execute(cql_lookup, [id_value, id_value2])
        ret = [i for i in ret]
        if len(ret) < 1:
            return None

        return ret[0]

    def truncate_table(self, model):
        cql = "TRUNCATE table {}.{}".format(
            model.__keyspace__,
            model.__table_name__
        )
        return self.session.execute(cql)








from .rq_raw_data_downloader import RqRawDataDownloader
from .em_raw_data_downloader import EmRawDataDownloader
from .web_raw_data_downloader import WebRawDataDownloader
from .raw_data_helper import RawDataHelper

class RawDataDownloader(object):
    def __init__(self, rq_license):
        self._data_helper = RawDataHelper()
        self.rq_downloader = RqRawDataDownloader(rq_license, self._data_helper)
        self.web_downloader = WebRawDataDownloader(self._data_helper)
        self.em_downloader = EmRawDataDownloader(self._data_helper)

    def download(self, start_date, end_date):
        failed_tasks = []

        failed_tasks.extend(self.em_downloader.download_all(start_date, end_date))
        # If 'em_tradedates' in failed_tasks, there is no trading day between start_date and end_date
        # Stop and return
        if 'em_tradedates' in failed_tasks:
            return failed_tasks

        failed_tasks.extend(self.rq_downloader.download_all(start_date, end_date))
        failed_tasks.extend(self.web_downloader.download_all(start_date, end_date))

        return failed_tasks

    def get_updated_count(self):
        return self._data_helper._updated_count




import dataclasses
import traceback
import datetime
from typing import List, Dict, Optional, Tuple
from functools import partial
from multiprocessing import Pool

import pandas as pd
import numpy as np

from ....util.calculator import Calculator
from ...api.raw import RawDataApi
from ...api.basic import BasicDataApi
from ...api.derived import DerivedDataApi
from ...view.derived_models import FundManagerIndex, FundManagerScore
from .derived_data_helper import DerivedDataHelper


class FundManagerIndexAndScore:

    _INDEX_LIST = {'csi_boodfund': 'H11023.CSI', 'csi_f_hybrid': 'H11022.CSI', 'csi_f_qdii': 'H11026.CSI', 'csi_stockfund': 'H11021.CSI', 'mmf': 'H11025.CSI'}
    _REPLACE_DICT = {'stock': 'H11021.CSI', 'bond': 'H11023.CSI', 'money': 'H11025.CSI', 'mix': 'H11022.CSI', 'QDII': 'H11026.CSI'}
    _DATE_DICT = {'stock': '2007-01-01', 'bond': '2003-01-01', 'money': '2004-01-07', 'mix': '2007-01-01', 'QDII': '2008-01-01'}

    def __init__(self):
        self._data_helper = DerivedDataHelper()
        self._raw_api = RawDataApi()
        self._basic_api = BasicDataApi()
        self._df_list: Dict[str, pd.DataFrame] = {}  # 基金经理指数
        self._rank_list: Dict[str, pd.DataFrame] = {}  # 基金经理评分

    def init(self, start_date: str, end_date: str):
        self._td: pd.DataFrame = self._basic_api.get_trading_day_list()
        self._td['datetime'] = self._td.datetime.map(str)
        print('load td done')

        self._nav: pd.DataFrame = self._raw_api.get_em_fund_nav(columns=('ADJUSTEDNAV',))
        self._nav = FundManagerIndexAndScore._clean_nav(self._nav, self._td)

        # 计算区间内的收益率
        self._rate: pd.DataFrame = self._nav.pct_change(fill_method=None).loc[start_date:end_date]
        # pd.testing.assert_index_equal(rate.index, self._scale.index)

        self._nav = self._nav.loc[start_date:end_date, :]
        print('load nav done')

        self._wfm: pd.DataFrame = self._raw_api.get_wind_manager_info(columns=('F_INFO_FUNDMANAGER', 'F_INFO_MANAGER_STARTDATE', 'F_INFO_MANAGER_LEAVEDATE'))
        self._wfm = FundManagerIndexAndScore._clean_wfm(self._wfm)
        print('load wfm done')

        self._scale: pd.DataFrame = self._raw_api.get_em_fund_scale()
        self._scale = FundManagerIndexAndScore._clean_scale(self._scale, self._td).loc[start_date:end_date].bfill()
        print('load scale done')

        self._base_index: pd.DataFrame = self._basic_api.get_index_price(index_list=list(self._INDEX_LIST.keys()))
        self._base_index = self._clean_index(self._base_index, self._td).loc[start_date:end_date]
        print('load base index done')

        self._info: pd.DataFrame = self._raw_api.get_wind_fund_info()
        print('load info done')

    @staticmethod
    def _gap_days(start: str, end: str) -> int:
        return (pd.to_datetime(end, infer_datetime_format=True) - pd.to_datetime(start, infer_datetime_format=True)).days

    @staticmethod
    def _filtered_by_class(info: pd.DataFrame, fund_kind: str, block_class_2_list: Tuple[str] = ()) -> pd.DataFrame:
        for one in block_class_2_list:
            info = info[~info['wind_class_2'].str.contains(one)]

        if fund_kind == 'QDII':
            info = info[info['wind_class_1'].str.contains('QDII')]
        elif fund_kind == 'money':
            info = info[info['wind_class_1'].str.contains('货币')]
        elif fund_kind == 'stock':
            info = info[info['wind_class_1'].str.contains('股') | (info['wind_class_2'].str.contains('股') & (~info['wind_class_2'].str.contains('QDII')))]
        elif fund_kind == 'bond':
            info = info[info['wind_class_1'].str.contains('债') | (info['wind_class_2'].str.contains('债') & (~info['wind_class_2'].str.contains('QDII')))]
        elif fund_kind == 'mix':
            info = info[info['wind_class_2'].str.contains('灵活') | info['wind_class_2'].str.contains('平衡')]
        return info

    @staticmethod
    def _clean_nav(nav: pd.DataFrame, td: pd.DataFrame) -> pd.DataFrame:
        nav.columns = ['fund_id', 'datetime', 'nav']
        nav['datetime'] = nav.datetime.map(str)
        nav = nav.pivot(index='datetime', columns='fund_id', values='nav')
        nav = nav.reindex(nav.index.union(td.datetime.array))
        nav = nav.ffill() + nav.bfill() - nav.bfill()  # bfill之后为NaN的部分让ffill之后也变为NaN
        nav = nav.loc[nav.index.intersection(td.datetime.array), :]
        return nav.dropna(how='all').dropna(how='all', axis=1)

    @staticmethod
    def _clean_wfm(wfm: pd.DataFrame) -> pd.DataFrame:
        wfm.columns = ['manager_id', 'fund_id', 'name', 'start', 'end']
        wfm['start'] = pd.to_datetime(wfm['start'], infer_datetime_format=True).dt.strftime('%Y-%m-%d').fillna('1962-01-01')
        wfm['end'] = pd.to_datetime(wfm['end'], infer_datetime_format=True).dt.strftime('%Y-%m-%d').fillna('2040-01-01')
        return wfm

    @staticmethod
    def _clean_scale(scale: pd.DataFrame, td: pd.DataFrame) -> pd.DataFrame:
        scale.columns = ['fund_id', 'datetime', 'scale']
        scale['datetime'] = scale.datetime.map(str)
        scale = scale.pivot(index='datetime', columns='fund_id', values='scale')
        scale = scale.reindex(scale.index.union(td.datetime.array)).ffill()
        scale = scale.loc[scale.index.intersection(td.datetime.array), :]
        return scale.dropna(how='all').dropna(how='all', axis=1)

    def _clean_index(self, index: pd.DataFrame, td: pd.DataFrame) -> pd.DataFrame:
        base_index = index.copy()
        base_index['index_id'] = base_index['index_id'].map(self._INDEX_LIST)
        base_index['datetime'] = base_index.datetime.map(str)
        base_index = base_index.pivot(index='datetime', columns='index_id', values='close')
        base_index = base_index.reindex(base_index.index.union(td.datetime.array))
        base_index = base_index.loc[base_index.index.intersection(td.datetime.array), :]
        return base_index.ffill().dropna(how='all').dropna(how='all', axis=1)

    @staticmethod
    def _lambda_func_1(single_fund_info: pd.DataFrame, x: pd.Series) -> pd.Series:
        result: List[pd.Series] = []
        for row in single_fund_info.itertuples(index=False):
            result.append(pd.Series((x.index >= row.start) & (x.index <= row.end), index=x.index))
        return pd.concat(result, axis=1).sum(axis=1) > 0

    def _index_calculation(self, manager: str, fund_kind: str, class_info: pd.DataFrame, rate_df: pd.DataFrame, index_rate: pd.DataFrame) -> Optional[pd.Series]:
        '''
        计算基金经理指数
        manager：基金经理的ID
        fund_kind：基金的分类（'stock', 'bond', 'money', 'mix', 'QDII'）
        replace_index：空窗期选择替代的指数的代码
        rate_df：基金日收益率的dataframe
        scale_df：基金规模的dataframe

        返回基金经理指数的Series，时间是从第一个产品的任职到最后一个产品的离职

        该指数由于bfill的原因不可回测
        replace_index为空时，空窗期按照_REPLACE_DICT中的指数来代替（如果没有指定基金的分类，按照沪深300的指数）
        基金规模按照上期季报的规模作为权重，如果没有之前的数据，则按照下期季报的规模作为权重
        如果该日基金净值缺失，则用之前的净值代替，如果历史上并没有该日的净值数据，那么在计算基金经理指数的过程中该基金产品不再拥有权重
        如果因该日基金净值缺失而导致该日基金经理没有基金产品可供计算指数，按照空窗期来处理
        '''
        m_info = class_info[class_info['manager_id'] == manager]
        if m_info.shape[0] == 0:
            # raise Exception('该经理没有相关基金产品')
            return
        m_start: str = m_info['start'].min()
        m_end: str = m_info['end'].max()
        unique_fund_ids = m_info['fund_id'].unique()
        rate: pd.DataFrame = rate_df.loc[m_start:m_end, unique_fund_ids]
        # 此指数不可回测
        scale: pd.DataFrame = self._scale.loc[m_start:m_end, unique_fund_ids]
        for fund_id in unique_fund_ids:
            single_fund_info: pd.DataFrame = m_info[m_info.fund_id == fund_id]
            rate.loc[:, fund_id] = rate[fund_id].where(partial(FundManagerIndexAndScore._lambda_func_1, single_fund_info), 0)
            scale.loc[:, fund_id] = scale[fund_id].where(partial(FundManagerIndexAndScore._lambda_func_1, single_fund_info), 0)
        scale = scale + rate - rate  # rate为NaN的地方也让scale变为NaN
        scale = scale.fillna(0)  # 这里在将NaN全变为0
        ratio = (rate * scale).sum(axis=1) / scale.sum(axis=1).replace(0, 1)
        if ratio.shape[0] == 0:
            return
        ratio = pd.Series(np.where(ratio, ratio, index_rate.loc[ratio.index[0]:ratio.index[-1]]), index=ratio.index).fillna(0)
        return np.exp(np.log(1 + ratio).cumsum())

    @staticmethod
    def _rate_calc(x: pd.Series) -> pd.Series:
        return x.pct_change(fill_method=None)

    def _calc_part3(self, eval_start_date: str, eval_date: str, m_dict, class_fund):
        _rank_list = {}
        for class_name, index_id in self._REPLACE_DICT.items():
            the_eval_start_date = max(eval_start_date, self._DATE_DICT[class_name])
            class_df = self._df_list[class_name]
            bm = self._base_index[index_id]
            try:
                cols = class_df.loc[eval_date].dropna().index
            except KeyError:
                continue
            stats: pd.DataFrame = pd.DataFrame.from_dict({'name': {i: m_dict[i] for i in cols}})
            benchmark_stat_result = []
            for i in cols:
                df = class_df[i].loc[the_eval_start_date:].dropna()
                benchmark = bm.loc[df.index]
                if benchmark.count() > 2:
                    # print(benchmark)
                    benchmark_stat_result.append(dataclasses.asdict(Calculator.get_benchmark_stat_result(df.index.array, df.array, benchmark.array)))
                else:
                    benchmark_stat_result.append({})
            stats = stats.assign(**(pd.DataFrame(benchmark_stat_result).to_dict(orient='list')))
            fund_ids = self._wfm['fund_id'].isin(class_fund[class_name])
            for row in stats.itertuples():
                stats.loc[row.Index, 'start_date'] = min(self._wfm[self._wfm['manager_id'].isin([row.Index]) & fund_ids]['start'])
                stats.loc[row.Index, 'trade_year'] = FundManagerIndexAndScore._gap_days(start=stats.loc[row.Index, 'start_date'], end=stats.loc[row.Index, 'end_date']) / 365
            rf = 1.025 ** (1 / 365) - 1
            cl_result = []
            for i in cols:
                df = class_df[i].loc[the_eval_start_date:].dropna()
                total = pd.concat([FundManagerIndexAndScore._rate_calc(bm.loc[df.index]) - rf, FundManagerIndexAndScore._rate_calc(df)], axis=1)
                total = total[total.notna().all(axis=1)]
                res = DerivedDataHelper._lambda_cl(total.to_numpy())
                cl_result.append(res)
            temp = pd.DataFrame(cl_result)
            stats = stats.assign(model_alpha=temp['alpha'].array, model_beta=temp['beta'].array)
            # 存进去原始数据而不是rank
            _rank_list[class_name] = stats

        print('score ranking...')
        for class_name in self._REPLACE_DICT.keys():
            try:
                stats = _rank_list[class_name].drop(columns=['name', 'start_date', 'end_date'])
                rank = stats.rank(pct=True) * 100
                if class_name != 'money':
                    rank['score'] = (rank['trade_year'] + rank['annualized_ret'] + rank['sharpe'] - rank['mdd'] + rank['model_alpha'] - rank['model_beta']).rank()
                else:
                    rank['score'] = (rank['trade_year'] + rank['annualized_ret'] - rank['annualized_vol']).rank()
                stats['score'] = 100 * rank['score'] / rank['score'].max()
                stats = stats[stats.notna().any(axis=1)]
                stats = stats.rename_axis(index='manager_id').reset_index()
                stats['fund_type'] = class_name
                stats['datetime'] = eval_date
                self._data_helper._upload_derived(stats, FundManagerScore.__table__.name)
            except KeyError:
                pass

    def calc(self, eval_date: str, eval_start_date: str, block_class_2_list: Tuple[str] = ()):
        print('to do part1')
        class_fund = {}
        class_index: Dict[str, pd.DataFrame] = {}
        class_info: Dict[str, pd.DataFrame] = {}
        m_dict = dict(zip(self._wfm['manager_id'].tolist(), self._wfm['name'].tolist()))
        # 在每个资产类别下用收益率的index定义一个空的df
        for class_name, index_id in self._REPLACE_DICT.items():
            self._df_list[class_name] = pd.DataFrame(index=self._rate.index)
            fund = FundManagerIndexAndScore._filtered_by_class(self._info, class_name, block_class_2_list)['wind_id']
            info = self._wfm.loc[self._wfm['fund_id'].isin(fund), :]
            info = info[info['fund_id'].isin(self._rate.columns.intersection(self._scale.columns))]
            class_info[class_name] = info
            class_index[class_name] = self._base_index[index_id].pct_change(fill_method=None)
            class_fund[class_name] = fund

        print('to do part2')
        import time
        _t0 = time.time()
        # 遍历每个基金经理
        for m_id in m_dict.keys():
            # 遍历每个资产类别
            for class_name in self._REPLACE_DICT.keys():
                idx = self._index_calculation(m_id, class_name, class_info[class_name], self._rate, class_index[class_name])
                if idx is not None:
                    # 该资产类别下记录计算结果
                    self._df_list[class_name].loc[:, m_id] = idx
        for class_name in self._REPLACE_DICT.keys():
            self._df_list[class_name] = self._df_list[class_name].replace(0, np.nan).dropna(how='all')
        _t1 = time.time()
        print(f'part2 costs {_t1 - _t0}s')

        print('to do part3')
        self._calc_part3(eval_start_date, eval_date, m_dict, class_fund)

    def _preprocess_manager_index_df(self, x):
        df = x.drop(columns='fund_type').pivot(index='datetime', columns='manager_id', values='manager_index')
        self._df_list[x.fund_type.array[0]] = df.set_axis(df.index.astype(str))

    def do_score_history(self, start_date: str, end_date: str):
        self._td: pd.DataFrame = self._basic_api.get_trading_day_list()
        self._td['datetime'] = self._td.datetime.map(str)
        print('load td done')

        self._wfm: pd.DataFrame = self._raw_api.get_wind_manager_info(columns=('F_INFO_FUNDMANAGER', 'F_INFO_MANAGER_STARTDATE', 'F_INFO_MANAGER_LEAVEDATE'))
        self._wfm = FundManagerIndexAndScore._clean_wfm(self._wfm)
        print('load wfm done')

        self._base_index: pd.DataFrame = self._basic_api.get_index_price(index_list=list(self._INDEX_LIST.keys()))
        self._base_index = self._clean_index(self._base_index, self._td).loc[start_date:end_date]
        print('load base index done')

        self._info: pd.DataFrame = self._raw_api.get_wind_fund_info()
        print('load info done')

        fund_manager_index: pd.DataFrame = DerivedDataApi().get_fund_manager_index().drop(columns='_update_time')
        print('load fund manager index done')
        fund_manager_index.groupby(by='fund_type', sort=False).apply(self._preprocess_manager_index_df)
        print('process fund manager index done')

        class_fund = {}
        m_dict = dict(zip(self._wfm['manager_id'].tolist(), self._wfm['name'].tolist()))
        for class_name, index_id in self._REPLACE_DICT.items():
            class_fund[class_name] = FundManagerIndexAndScore._filtered_by_class(self._info, class_name)['wind_id']
        print('all data done, to calc fund manager score day by day')

        min_date = fund_manager_index.datetime.min()
        for eval_date in self._td.datetime.array:
            if eval_date < '2011-04-26':
                continue
            eval_date_dt = pd.to_datetime(eval_date, infer_datetime_format=True).date()
            if eval_date_dt < min_date:
                continue
            eval_start_date_dt = eval_date_dt - datetime.timedelta(days=365)
            self._calc_part3(eval_start_date_dt.isoformat(), eval_date_dt.isoformat(), m_dict, class_fund)
            print(f'{eval_date_dt} done')

        # p = Pool()
        # p.imap_unordered(partial(self._calc_part3, m_dict, class_fund), self.fund_list, 256)
        # p.close()
        # p.join()

    def process(self, start_date: str, end_date: str, eval_start_date: str, eval_date: str) -> List[str]:
        failed_tasks: List[str] = []
        try:
            self.init(start_date, end_date)
            print('init done, to calc manager index and score')
            # 制作二个字典，键均为基金类型，值分别为基金经理指数以及基金经理得分(包括计算得分的因子)
            self.calc(eval_date, eval_start_date)
            for class_type, df in self._df_list.items():
                index_to_inserted = df.rename_axis(index='datetime', columns='manager_id').stack().rename('manager_index').reset_index()
                index_to_inserted['fund_type'] = class_type
                self._data_helper._upload_derived(index_to_inserted, FundManagerIndex.__table__.name)

            # for class_type, df in self._rank_list.items():
            #     score_to_inserted = df.drop(columns='name').rename(columns={'end_date': 'datetime'}).rename_axis(index='manager_id').reset_index()
            #     score_to_inserted['fund_type'] = class_type
            #     self._data_helper._upload_derived(score_to_inserted, FundManagerScore.__table__.name)
        except Exception as e:
            print(e)
            traceback.print_exc()
            failed_tasks.append('fund_manager_index_and_score')
        return failed_tasks


if __name__ == '__main__':
    fmis = FundManagerIndexAndScore()
    fmis.process(start_date='2005-01-01', end_date='2020-09-07', eval_date='2020-09-07', eval_start_date='2005-01-01')
    # fmis.do_score_history(start_date='2005-01-01', end_date='2020-09-04')




import pandas as pd
import numpy as np
import datetime
import traceback
import time
from typing import List, Dict, Optional
from functools import partial
from multiprocessing import Pool

from ....util.calculator import Calculator
from ...api.basic import BasicDataApi
from ...api.derived import DerivedDataApi
from ...wrapper.mysql import BasicDatabaseConnector, DerivedDatabaseConnector
from ...view.basic_models import FundNav, Fund_size_and_hold_rate, FundInfo
from ...view.derived_models import FundManagerInfo, FundManagerIndex, FundManagerScore
from .derived_data_helper import DerivedDataHelper, FUND_CLASSIFIER, normalize, score_rescale


class ManagerProcessor:

    INDEX_LIST = {'csi_boodfund': 'H11023.CSI', 'csi_f_hybrid': 'H11022.CSI', 'csi_f_qdii': 'H11026.CSI', 'csi_stockfund': 'H11021.CSI', 'mmf': 'H11025.CSI'}
    REPLACE_DICT = {'stock': 'H11021.CSI', 'bond': 'H11023.CSI', 'mmf': 'H11025.CSI', 'QDII': 'H11026.CSI', 'mix': 'H11022.CSI'}
    DATE_DICT = {'stock': '2007-01-01', 'bond': '2003-01-01', 'mmf': '2004-01-07', 'QDII': '2008-01-01', 'mix': '2007-01-01'}

    def __init__(self, data_helper: DerivedDataHelper):
        self._data_helper: DerivedDataHelper = data_helper
        self._basic_api = BasicDataApi()
        self._df_list: Dict[str, pd.DataFrame] = {}  # 基金经理指数
        self._rank_list: Dict[str, pd.DataFrame] = {}  # 基金经理评分

    def init(self, start_date=str, end_date=str):
        self._td: pd.DataFrame = self._basic_api.get_trading_day_list()
        self._td['datetime'] = self._td.datetime.map(str)
        self._td = self._td[(self._td.datetime >= start_date) & (self._td.datetime <= end_date)]
        print('load td done')

        with BasicDatabaseConnector().managed_session() as quant_session:
            fund_nav_query = quant_session.query(
                FundNav.fund_id,
                FundNav.adjusted_net_value,
                FundNav.datetime
            ).filter(
                FundNav.datetime >= start_date,
                FundNav.datetime <= end_date,
            )
            self._nav = pd.read_sql(fund_nav_query.statement, fund_nav_query.session.bind)
            self._nav = ManagerProcessor._clean_nav(self._nav, self._td)
            self._rate: pd.DataFrame = self._nav.pct_change(fill_method=None).loc[start_date:end_date]
            print('load nav done')

        with DerivedDatabaseConnector().managed_session() as quant_session:
            mng_query = quant_session.query(
                FundManagerInfo.mng_id,
                FundManagerInfo.start_date,
                FundManagerInfo.fund_id,
                FundManagerInfo.mng_name,
                FundManagerInfo.end_date)
            self._fm: pd.DataFrame = pd.read_sql(mng_query.statement, mng_query.session.bind)
            self._fm = ManagerProcessor._clean_fm(self._fm)
            print('load fm done')

        with BasicDatabaseConnector().managed_session() as quant_session:
            fund_size_query = quant_session.query(
                Fund_size_and_hold_rate.fund_id,
                Fund_size_and_hold_rate.datetime,
                Fund_size_and_hold_rate.size
            )
            self._scale: pd.DataFrame = pd.read_sql(fund_size_query.statement, fund_size_query.session.bind)
            self._scale = ManagerProcessor._clean_scale(self._scale, self._td).loc[start_date:end_date].bfill()
            print('load scale done')

        self._base_index: pd.DataFrame = self._basic_api.get_index_price(index_list=list(self.INDEX_LIST.keys()))
        self._base_index = self._clean_index(self._base_index, self._td).loc[start_date:end_date]
        print('load base index done')

        with BasicDatabaseConnector().managed_session() as quant_session:
            fund_info_query = quant_session.query(FundInfo)
        self._info: pd.DataFrame = pd.read_sql(fund_info_query.statement, fund_info_query.session.bind)
        print('load info done')

    @staticmethod
    def _gap_days(start: str, end: str) -> int:
        return (pd.to_datetime(end, infer_datetime_format=True) - pd.to_datetime(start, infer_datetime_format=True)).days

    @staticmethod
    def _rate_calc(x: pd.Series) -> pd.Series:
        return x.pct_change(fill_method=None)

    @staticmethod
    def _clean_nav(nav: pd.DataFrame, td: pd.DataFrame) -> pd.DataFrame:
        nav = nav.rename(columns={'adjusted_net_value': 'nav'})
        nav['datetime'] = nav.datetime.map(str)
        nav = nav.pivot(index='datetime', columns='fund_id', values='nav')
        nav = nav.reindex(nav.index.union(td.datetime.array))
        nav = nav.ffill() + nav.bfill() - nav.bfill()  # bfill之后为NaN的部分让ffill之后也变为NaN
        nav = nav.loc[nav.index.intersection(td.datetime.array), :]
        return nav.dropna(how='all').dropna(how='all', axis=1)

    @staticmethod
    def _clean_fm(fm: pd.DataFrame) -> pd.DataFrame:
        fm = fm.rename(columns={'mng_id': 'manager_id', 'mng_name': 'name', 'start_date': 'start', 'end_date': 'end'})
        fm['start'] = pd.to_datetime(fm.start).dt.strftime('%Y-%m-%d')
        fm['end'] = pd.to_datetime(fm.end).dt.strftime('%Y-%m-%d')
        fm = fm[fm.fund_id != 'not_exist'].copy()
        return fm

    @staticmethod
    def _clean_scale(scale: pd.DataFrame, td: pd.DataFrame) -> pd.DataFrame:
        scale = scale.rename(columns={'size': 'scale'})
        scale['datetime'] = scale.datetime.map(str)
        scale = scale.pivot(index='datetime', columns='fund_id', values='scale')
        scale = scale.reindex(scale.index.union(td.datetime.array)).ffill()
        scale = scale.loc[scale.index.intersection(td.datetime.array), :]
        return scale.dropna(how='all').dropna(how='all', axis=1)

    def _clean_index(self, index: pd.DataFrame, td: pd.DataFrame) -> pd.DataFrame:
        base_index = index.copy()
        base_index['index_id'] = base_index['index_id'].map(self.INDEX_LIST)
        base_index['datetime'] = base_index.datetime.map(str)
        base_index = base_index.pivot(index='datetime', columns='index_id', values='close')
        base_index = base_index.reindex(base_index.index.union(td.datetime.array))
        base_index = base_index.loc[base_index.index.intersection(td.datetime.array), :]
        return base_index.ffill().dropna(how='all').dropna(how='all', axis=1)

    @staticmethod
    def _lambda_func_1(single_fund_info: pd.DataFrame, x: pd.Series) -> pd.Series:
        result: List[pd.Series] = []
        for row in single_fund_info.itertuples(index=False):
            result.append(pd.Series((x.index >= row.start) & (x.index <= row.end), index=x.index))
        return pd.concat(result, axis=1).sum(axis=1) > 0

    def _index_calculation(self, manager: str, fund_kind: str, class_info: pd.DataFrame, rate_df: pd.DataFrame, index_rate: pd.DataFrame) -> Optional[pd.Series]:
        '''
        计算基金经理指数
        manager：基金经理的ID
        fund_kind：基金的分类（'stock', 'bond', 'money', 'mix', 'QDII'）
        replace_index：空窗期选择替代的指数的代码
        rate_df：基金日收益率的dataframe
        scale_df：基金规模的dataframe

        返回基金经理指数的Series，时间是从第一个产品的任职到最后一个产品的离职

        该指数由于bfill的原因不可回测
        replace_index为空时，空窗期按照_REPLACE_DICT中的指数来代替（如果没有指定基金的分类，按照沪深300的指数）
        基金规模按照上期季报的规模作为权重，如果没有之前的数据，则按照下期季报的规模作为权重
        如果该日基金净值缺失，则用之前的净值代替，如果历史上并没有该日的净值数据，那么在计算基金经理指数的过程中该基金产品不再拥有权重
        如果因该日基金净值缺失而导致该日基金经理没有基金产品可供计算指数，按照空窗期来处理
        '''
        m_info = class_info[class_info['manager_id'] == manager]
        if m_info.shape[0] == 0:
            # raise Exception('该经理没有相关基金产品')
            return
        m_start: str = m_info['start'].min()
        m_end: str = m_info['end'].max()
        unique_fund_ids = m_info['fund_id'].unique()
        rate: pd.DataFrame = rate_df.loc[m_start:m_end, unique_fund_ids]
        # 此指数不可回测
        scale: pd.DataFrame = self._scale.loc[m_start:m_end, unique_fund_ids]
        for fund_id in unique_fund_ids:
            single_fund_info: pd.DataFrame = m_info[m_info.fund_id == fund_id]
            rate.loc[:, fund_id] = rate[fund_id].where(partial(ManagerProcessor._lambda_func_1, single_fund_info), 0)
            scale.loc[:, fund_id] = scale[fund_id].where(partial(ManagerProcessor._lambda_func_1, single_fund_info), 0)
        scale = scale + rate - rate  # rate为NaN的地方也让scale变为NaN
        scale = scale.fillna(0)  # 这里在将NaN全变为0
        ratio = (rate * scale).sum(axis=1) / scale.sum(axis=1).replace(0, 1)
        if ratio.shape[0] == 0:
            return
        ratio = pd.Series(np.where(ratio, ratio, index_rate.loc[ratio.index[0]:ratio.index[-1]]), index=ratio.index).fillna(0)
        return np.exp(np.log(1 + ratio).cumsum())

    def _calc_part3(self, m_dict, class_fund, eval_date_dt):
        eval_date = eval_date_dt.isoformat()
        eval_start_date = (eval_date_dt - datetime.timedelta(days=365)).isoformat()
        self._do_calc_part3(eval_start_date, eval_date, m_dict, class_fund)

    def _do_calc_part3(self, eval_start_date: str, eval_date: str, m_dict, class_fund):
        _rank_list = {}
        _tm_part3_start = time.time()
        for class_name, index_id in self.REPLACE_DICT.items():
            the_eval_start_date = max(eval_start_date, self.DATE_DICT[class_name])
            class_df = self._df_list[class_name]
            bm = self._base_index[index_id]
            try:
                cols = class_df.loc[eval_date].dropna().index
            except KeyError:
                continue
            benchmark_stat_result = []
            for i in cols:
                df = class_df[i].loc[the_eval_start_date:].dropna()
                if df.shape[0] < 5:
                    continue
                dic = Calculator.get_manager_stat_result(df.index.array, df.array, bm.loc[df.index].array)
                dic['name'] = m_dict[i]
                dic['index'] = i
                benchmark_stat_result.append(dic)
            stats: pd.DataFrame = pd.DataFrame(benchmark_stat_result).set_index('index')
            fund_ids = self._fm['fund_id'].isin(class_fund[class_name])
            for row in stats.itertuples():
                stats.loc[row.Index, 'start_date'] = min(self._fm[self._fm['manager_id'].isin([row.Index]) & fund_ids]['start'])
                stats.loc[row.Index, 'trade_year'] = ManagerProcessor._gap_days(start=stats.loc[row.Index, 'start_date'], end=stats.loc[row.Index, 'end_date']) / 365
            rf = 1.025 ** (1 / 365) - 1
            cl_result = []
            for i in stats.index:
                df = class_df[i].loc[the_eval_start_date:].dropna()
                total = pd.concat([ManagerProcessor._rate_calc(bm.loc[df.index]) - rf, ManagerProcessor._rate_calc(df)], axis=1)
                total = total[total.notna().all(axis=1)]
                res = DerivedDataHelper._lambda_cl(total.to_numpy())
                cl_result.append(res)
            temp = pd.DataFrame(cl_result)
            stats = stats.assign(model_alpha=temp['alpha'].array, model_beta=temp['beta'].array)
            # 存进去原始数据而不是rank
            _rank_list[class_name] = stats
        _tm_part3_end = time.time()
        print(f' part3 cost time {_tm_part3_end - _tm_part3_start}')

        _tm_part4_start = time.time()
        for class_name in self.REPLACE_DICT.keys():
            try:
                stats = _rank_list[class_name].drop(columns=['name', 'start_date', 'end_date'])
                if class_name != 'money':
                    stats['score'] = score_rescale(normalize(stats['trade_year']) + normalize(stats['annualized_ret']) + normalize(stats['sharpe']) + normalize(-stats['mdd']) + normalize(stats['model_alpha']) + normalize(stats['model_beta']))
                else:
                    stats['score'] = score_rescale(normalize(stats['trade_year']) + normalize(stats['annualized_ret']) + normalize(-stats['annualized_vol']))
                stats = stats[stats.notna().any(axis=1)]
                stats = stats.rename_axis(index='manager_id').reset_index()
                stats['fund_type'] = class_name
                stats['datetime'] = eval_date
                self._data_helper._upload_derived(stats, FundManagerScore.__table__.name)
            except KeyError:
                pass
        _tm_part4_end = time.time()
        print(f' part4 cost time {_tm_part4_end - _tm_part4_start}')
        print(f'{eval_date} done, cost {_tm_part4_end - _tm_part3_start}')

    def calc(self):
        _tm_part1_start = time.time()
        class_fund = {}
        class_index: Dict[str, pd.DataFrame] = {}
        class_info: Dict[str, pd.DataFrame] = {}
        m_dict = dict(zip(self._fm['manager_id'].tolist(), self._fm['name'].tolist()))
        # 在每个资产类别下用收益率的index定义一个空的df
        for class_name, index_id in self.REPLACE_DICT.items():
            self._df_list[class_name] = pd.DataFrame(index=self._rate.index)
            fund = self._info[self._info['wind_class_2'].isin(FUND_CLASSIFIER[class_name])].fund_id
            info = self._fm.loc[self._fm['fund_id'].isin(fund), :]
            info = info[info['fund_id'].isin(self._rate.columns.intersection(self._scale.columns))]
            class_info[class_name] = info
            class_index[class_name] = self._base_index[index_id].pct_change(fill_method=None)
            class_fund[class_name] = fund
        _tm_part1_end = time.time()
        print(f' part1 cost time {_tm_part1_end - _tm_part1_start}')

        # 遍历每个基金经理
        _tm_part2_start = time.time()
        for m_id in m_dict.keys():
            # 遍历每个资产类别
            for class_name in self.REPLACE_DICT.keys():
                idx = self._index_calculation(m_id, class_name, class_info[class_name], self._rate, class_index[class_name])
                if idx is not None:
                    # 该资产类别下记录计算结果
                    self._df_list[class_name].loc[:, m_id] = idx
        for class_name in self.REPLACE_DICT.keys():
            self._df_list[class_name] = self._df_list[class_name].replace(0, np.nan).dropna(how='all')
        _tm_part2_end = time.time()
        print(f' part2 cost time {_tm_part2_end - _tm_part2_start}')

    def _preprocess_manager_index_df(self, x):
        df = x.drop(columns='fund_type').pivot(index='datetime', columns='manager_id', values='manager_index')
        self._df_list[x.fund_type.array[0]] = df.set_axis(df.index.astype(str))

    def do_score_history(self, start_date: str, end_date: str):
        self._td: pd.DataFrame = self._basic_api.get_trading_day_list()
        self._td['datetime'] = self._td.datetime.map(str)
        self._td = self._td[(self._td.datetime >= start_date) & (self._td.datetime <= end_date)]
        print('load td done')

        with DerivedDatabaseConnector().managed_session() as quant_session:
            mng_query = quant_session.query(
                FundManagerInfo.mng_id,
                FundManagerInfo.start_date,
                FundManagerInfo.fund_id,
                FundManagerInfo.mng_name,
                FundManagerInfo.end_date)
            self._fm: pd.DataFrame = pd.read_sql(mng_query.statement, mng_query.session.bind)
            self._fm = ManagerProcessor._clean_fm(self._fm)
            print('load fm done')

        self._base_index: pd.DataFrame = self._basic_api.get_index_price(index_list=list(self.INDEX_LIST.keys()))
        self._base_index = self._clean_index(self._base_index, self._td).loc[start_date:end_date]
        print('load base index done')

        with BasicDatabaseConnector().managed_session() as quant_session:
            fund_info_query = quant_session.query(FundInfo)
        self._info: pd.DataFrame = pd.read_sql(fund_info_query.statement, fund_info_query.session.bind)
        print('load info done')

        fund_manager_index: pd.DataFrame = DerivedDataApi().get_fund_manager_index().drop(columns='_update_time')
        print('load fund manager index done')
        fund_manager_index.groupby(by='fund_type', sort=False).apply(self._preprocess_manager_index_df)
        print('process fund manager index done')

        m_dict = dict(zip(self._fm['manager_id'].tolist(), self._fm['name'].tolist()))
        class_fund = {}
        for class_name, index_id in self.REPLACE_DICT.items():
            class_fund[class_name] = self._info[self._info['wind_class_2'].isin(FUND_CLASSIFIER[class_name])].fund_id
        print('all data done, to calc fund manager score day by day')

        eval_date_dt_list: List[datetime.date] = []
        min_date = fund_manager_index.datetime.min()
        for eval_date in self._td.datetime.sort_values(ascending=False).array:
            eval_date_dt = pd.to_datetime(eval_date, infer_datetime_format=True).date()
            if eval_date_dt < min_date:
                continue

            eval_date_dt_list.append(eval_date_dt)
            # eval_start_date_dt = eval_date_dt - datetime.timedelta(days=365)
            # self._calc_part3(eval_start_date_dt.isoformat(), eval_date_dt.isoformat(), m_dict, class_fund)
            # print(f'{eval_date_dt} done')

        p = Pool()
        for i in p.imap_unordered(partial(self._calc_part3, m_dict, class_fund), eval_date_dt_list, 16):
            pass
        print('all done, to close and join')
        p.close()
        p.join()

    def do_score_update(self, eval_date_dt: datetime.date):
        m_dict = dict(zip(self._fm['manager_id'].tolist(), self._fm['name'].tolist()))
        class_fund = {}
        for class_name, index_id in self.REPLACE_DICT.items():
            class_fund[class_name] = self._info[self._info['wind_class_2'].isin(FUND_CLASSIFIER[class_name])].fund_id
        self._calc_part3(m_dict, class_fund, eval_date_dt)

    def do_index_history(self, start_date: str, end_date: str, history: bool = False) -> List[str]:
        self.init(start_date, end_date)
        print('init done, to calc manager index')
        self.calc()
        for class_type, df in self._df_list.items():
            index_to_inserted = df.rename_axis(index='datetime', columns='manager_id').stack().rename('manager_index').reset_index()
            if not history:
                index_to_inserted = index_to_inserted[index_to_inserted.datetime == end_date]
            index_to_inserted['fund_type'] = class_type
            self._data_helper._upload_derived(index_to_inserted, FundManagerIndex.__table__.name)

    def process(self, end_date: str):
        failed_tasks = []
        try:
            end_date_dt: datetime.date = pd.to_datetime(end_date, infer_datetime_format=True).date()
            self.do_index_history('2005-01-01', end_date_dt.isoformat())
            self.do_score_update(end_date_dt)
        except Exception as e:
            print(e)
            traceback.print_exc()
            failed_tasks.append('fund_manager_processor')
        return failed_tasks


if __name__ == '__main__':
    fmis = ManagerProcessor(DerivedDataHelper())
    # fmis.process('2020-09-10')
    # fmis.do_index_history(start_date='2005-01-01', end_date='2020-09-07', eval_date='2020-09-07', eval_start_date='2005-01-01')
    # fmis.do_score_history(start_date='2005-01-01', end_date='2020-09-07')




import pickle
import os
import gzip
import datetime
import time
from ..fund.engine.backtest import *


class DataStore:

    REPORT_PATH = 'report_data.pkz'
    DATA_MANAGER_PATH = 'data_manager.pkz'
    CURR_FILE_DIR = os.path.split(os.path.realpath(__file__))[0]

    def __init__(self):
        pass

    @staticmethod
    def store(data, f_path):
        f_path = os.path.join(DataStore.CURR_FILE_DIR, f_path)
        with gzip.GzipFile(f_path, 'wb') as f:
            pickle.dump(data, f, -1)
        print(f'store data successfully to {f_path}')

    @staticmethod
    def load(f_path):
        f_path = os.path.join(DataStore.CURR_FILE_DIR, f_path)
        with gzip.GzipFile(f_path, 'rb') as f:
            data = pickle.load(f)
        print(f'load data successfully from {f_path}')
        return data

    @staticmethod
    def refresh_dm(score_pre_calc=True,start_time='20050101',print_time=False,use_weekly_monthly_indicators=False,s3_retriever_activation=False):
        print(f'start load at {datetime.datetime.now()}')
        t1 = time.time()
        dm = FundDataManager(start_time=start_time, end_time=datetime.datetime.now().strftime('%Y%m%d'), s3_retriever_activation=s3_retriever_activation)
        dm.init(score_pre_calc=score_pre_calc, print_time=print_time, use_weekly_monthly_indicators=use_weekly_monthly_indicators)
        DataStore.store(dm, DataStore.DATA_MANAGER_PATH)
        t2 = time.time()
        print(f'refresh dm cost time {t2 - t1}s')

    @staticmethod
    def load_dm():
        return DataStore.load(DataStore.DATA_MANAGER_PATH)

    @staticmethod
    def save_reporter(reporter):
        DataStore.store(reporter, DataStore.REPORT_PATH)

    @staticmethod
    def load_reporter():
        return DataStore.load(DataStore.REPORT_PATH)

if __name__ == "__main__":
    # save fm
    DataStore.refresh_dm()
    # load dm
    m = DataStore.load_dm()
    taa_detail_item_mmf = TAAParam(HighThreshold = 0.95,
                   HighStop = 0.5,
                   HighMinus = 0.03,
                   LowStop = -1,
                   LowThreshold = -1,
                   LowPlus = 0,
                   ToMmf=True)

    taa_detail= {
        'hs300': taa_detail_item_mmf,
        'csi500': taa_detail_item_mmf,
        'gem': taa_detail_item_mmf,
        'sp500rmb': taa_detail_item_mmf,
    }
    asset_param = AssetTradeParam()
    fund_param = FundTradeParam()
    t = FundTrader(asset_param, fund_param)
    fa_param = FAParam()
    taa_default = TAAParam()
    saa = AssetWeight(
        hs300=18/100,
        csi500=14/100,
        gem=9/100,
        sp500rmb=4/100,
        national_debt=44/100,
        gold=10/100,
        cash=1/100
    )
    b = FundBacktestEngine(data_manager=m, trader=t, taa_params=taa_default, fa_params=fa_param, taa_param_details=taa_detail)
    b.init()
    b.run(saa=saa, start_date=datetime.date(2020,1,3), end_date=datetime.date(2020,6,4))
    # save reporter
    DataStore.save_reporter(b._report_helper)

    # load reporter
    reporter = DataStore.load_reporter()


import pandas as pd
import numpy as np
import datetime
import matplotlib as mpl
import pylab as pl
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import matplotlib.colors as colors
mpl.rcParams['font.family'] = ['Heiti TC']

from .manager_fund import FundDataManager
from .score import FundScoreManager
from ..struct import AssetWeight

class FundIndicatorCheck:

    def __init__(self, start_time=None, end_time=None):
        self._dm = FundDataManager(start_time=start_time, end_time=end_time, score_manager=FundScoreManager())
        self.fund_score = FundScoreManager()

    def init(self, asset, check_date):
        self._dm.init()
        self.asset = asset
        self.check_date = check_date
        
        self.tradings_days = self._dm.dts.trading_days
        self.fund_info = self._dm.dts.fund_info
        self.fund_indicator = self._dm.dts.fund_indicator.pivot_table(index=['fund_id','datetime'])
        self.fund_nav = self._dm.dts.fund_nav
        self.index_price = self._dm.dts.index_price
        self.score_equation = self.fund_score.funcs[self.asset].__dict__

        self.check_date = self.tradings_days[self.tradings_days.datetime > self.check_date].datetime.values[0]
        self.fund_list = self.fund_info[self.fund_info['index_id'].isin([self.asset])].fund_id.tolist()
        self.fund_list = [_ for _ in self.fund_nav.columns if _ in self.fund_list]
        self.fund_nav = self.fund_nav[self.fund_list]
        self.index_price = self.index_price[[asset]].loc[:check_date,]
        self.fund_indicator = self.fund_indicator.loc[(self.fund_list,self.check_date),]

    def check_indicator(self):
        res = []
        for fund_id in self.fund_nav.columns:
            dftmp = self.fund_nav[fund_id]
            res.append({'fund_id':fund_id,'begin_nav_date':dftmp.dropna().index[0]})
        res = pd.DataFrame(res).set_index('fund_id')
        df = res.join(self.fund_info.set_index('fund_id'))[['begin_nav_date','desc_name']].join(self.fund_indicator).sort_values('begin_nav_date')
        return df

    def plot_price_ratio(self, select_fund_id:list):
        df = self.fund_nav[select_fund_id].join(self.index_price).dropna()
        df = df / df.iloc[0]
        df.plot.line(figsize=(16,8),legend=False,fontsize = 13)
        l = pl.legend(loc='lower left',fontsize = 10)
        s = pl.title('price ratio', fontsize=20)

if __name__ == "__main__":
    checker = FundIndicatorCheck(start_time = '20100101', end_time = '20160101')
    checker.init(asset='gem', check_date=datetime.date(2015,1,6))
    checker.check_indicator()
    checker.plot_price_ratio(select_fund_id = ['050021!0','159908!0'])


from .compute_fund_daily_collection import FundDailyCollectionProcessor
from .compute_index_daily_collection import IndexDailyCollectionProcessor
from .compute_aic_daily_collection import AICDailyCollectionProcessor

class ViewDataProcessor(object):
    def __init__(self):
        self.fund_daily_collection_processor = FundDailyCollectionProcessor()
        self.index_daily_collection_processor = IndexDailyCollectionProcessor()
        self.aic_daily_collection_processor = AICDailyCollectionProcessor()

    def process_all(self, start_date, end_date):
        failed_tasks = []

        failed_tasks.extend(self.fund_daily_collection_processor.process())
        failed_tasks.extend(self.index_daily_collection_processor.process())
        failed_tasks.extend(self.aic_daily_collection_processor.process())

        return failed_tasks



from cassandra.cqlengine import columns
from cassandra.cqlengine.models import Model
from ...wrapper.cas.cas_connection import CassandraConnector
from ...wrapper.cas.constants import CAS_KEY_SPACE

CassandraConnector().get_conn()

class RealtimeIndexPriceSnapshot(Model):
    __table_name__ = 'realtime_index_price_snapshot'
    __keyspace__ = CAS_KEY_SPACE
    __connection__ = 'surfing'

    ts = columns.DateTime(primary_key=True)                   # 数据时间
    index_prices = columns.Map(columns.Text, columns.Double)         # 指数点位


    def to_dict(self):
        return {
            'ts': self.ts,
            'index_prices': self.index_prices
        }






import datetime
import traceback
import pandas as pd
from cassandra.cqlengine.query import DoesNotExist
from cassandra.cqlengine.query import BatchQuery

from ...api.raw import RawDataApi
from ....constant import IndexPriceSource
from ...api.basic import BasicDataApi
from ...view.cas.realtime_index_price import RealtimeIndexPrice
from ...view.cas.realtime_index_price_snapshot import RealtimeIndexPriceSnapshot


class RealtimeIndexPriceProcessor(object):

    def init(self):
        try:
            today = datetime.datetime.today().strftime('%Y-%m-%d')
            trading_day_list = BasicDataApi().get_trading_day_list(start_date=today, end_date=today)
            if trading_day_list is None or trading_day_list.empty:
                print(f'Today {today} is not trading day!')
                return False
                
            last_trading_day = BasicDataApi().get_last_trading_day()
            print(f'last_trading_day: {last_trading_day}')
            if not last_trading_day:
                print(f'Cannot find last trading day!')
                return False

            index_info_df = BasicDataApi().get_index_info()
            if index_info_df is None:
                print('Failed to get index list')
                return False
            index_info_df = index_info_df[index_info_df.price_source == IndexPriceSource.default]
            self.index_ids = set(index_info_df.em_id.to_list())
            self.index_ids.discard(None)

            self.index_id_mapping = self.load_index_id_mapping(index_info_df)

            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def load_index_id_mapping(self, index_info_df):
        index_id_mapping = {}
        for row in index_info_df.itertuples(index=False):
            if not row.em_id:
                continue
            index_id_mapping[row.em_id] = row.index_id
        return index_id_mapping

    def load_realtime_index_price(self, realtime_index_price_df):
        if realtime_index_price_df is None or realtime_index_price_df.empty:
            return None

        realtime_index_price = {}
        for row in realtime_index_price_df.itertuples(index=False):
            realtime_index_price[row.em_id] = row.price
        return realtime_index_price

    def calc_realtime_index_price(self, realtime_index_price_df, curr_datetime):
        try:
            realtime_index_price = self.load_realtime_index_price(realtime_index_price_df)
            if not realtime_index_price:
                return False

            results = []
            results_snapshot = {}
            for em_id, price in realtime_index_price.items():
                if em_id not in self.index_id_mapping:
                    continue

                results.append([self.index_id_mapping[em_id], price, curr_datetime])
                results_snapshot[self.index_id_mapping[em_id]] = price

                if len(results) >= 200:
                    # Write Cassandra DB
                    with BatchQuery() as batch:
                        for res in results:
                            RealtimeIndexPrice.objects(index_id=res[0]).batch(batch).update(
                                prices__append = [res[1]],
                                timestamps__append = [res[2]],
                            )
                    results = []
            
            if results:
                # Write Cassandra DB
                with BatchQuery() as batch:
                    for res in results:
                        RealtimeIndexPrice.objects(index_id=res[0]).batch(batch).update(
                            prices__append = [res[1]],
                            timestamps__append = [res[2]],
                        )

            RealtimeIndexPriceSnapshot().create(
                ts = curr_datetime,
                index_prices = results_snapshot
            )

        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def dump_index_list(self):
        '''
        Utility function
        '''
        import time
        import datetime
        import pandas as pd

        from ....util.config import SurfingConfigurator
        from ...view.raw_models import EmRealtimeIndexPrice
        from ...wrapper.mysql import RawDatabaseConnector
        from ...api.basic import BasicDataApi
        from ..raw.raw_data_helper import RawDataHelper
        from ..raw.em_raw_data_downloader import EmRawDataDownloader

        with RawDatabaseConnector().managed_session() as db_session:
            query = db_session.query(
                EmRealtimeIndexPrice.em_id
            )
            df = pd.read_sql(query.statement, query.session.bind)
        print(df)
        index_ids = set(df.em_id.to_list())
        print(index_ids)

        index_info_df = BasicDataApi().get_index_info().reset_index()
        index_id_mapping = {}
        realtime = pd.DataFrame(columns=['index_id', 'desc_name', 'is_select', 'em_id'])
        count_realtime = 0
        others = pd.DataFrame(columns=['index_id', 'desc_name', 'is_select', 'em_id'])
        count_others = 0
        for row in index_info_df.itertuples(index=False):
            if row.em_id in index_ids:
                realtime.loc[count_realtime] = [row.index_id, row.desc_name, row.is_select, row.em_id]
                count_realtime += 1
            else:
                others.loc[count_others] = [row.index_id, row.desc_name, row.is_select, row.em_id]
                count_others += 1

        realtime.to_excel('./realtime_index.xlsx')
        others.to_excel('./other_index.xlsx')